from __future__ import unicode_literals
from nose.plugins.attrib import attr
from nose.tools import nottest
from mogwai.tests import BaseMogwaiTestCase
from mogwai._compat import PY2
from .base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import DateTimeNaive, GraphProperty
from mogwai.models import Vertex
from mogwai.exceptions import ValidationError
from mogwai._compat import print_
import datetime


@attr('unit', 'property', 'property_datetime')
class DateTimeNaivePropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = DateTimeNaive
    good_cases = (datetime.datetime.now(), None)
    if PY2:
        bad_cases = ('val', [], (), {}, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), {}, 0, False, 1.1)

    def test_to_database_method(self):
        d = self.klass(strict=False)
        self.assertIsNone(d.to_database(None))
        self.assertIsInstance(d.to_database(100000), float)
        with self.assertRaises(ValidationError):
            d.to_database(lambda x: x)

    def test_input_output_equality(self):
        now = datetime.datetime.now()
        prop = self.klass()
        self.assertAlmostEqual(now, prop.to_python(prop.to_database(now)), delta=datetime.timedelta(milliseconds=1))


class DateTimeNaiveTestVertex(Vertex):
    element_type = 'test_datetime_naive_vertex'

    test_val = DateTimeNaive()

CHOICES = (
    (datetime.datetime(2014, 1, 1), 'A'),
    (datetime.datetime(2014, 2, 1), 'B')
)


class DateTimeNaiveTestChoicesVertex(Vertex):
    element_type = 'test_datetime_naive_choices_vertex'

    test_val = DateTimeNaive(choices=CHOICES)


@attr('unit', 'property', 'property_datetime')
class DateTimeNaiveVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_datetime_naive_io(self):
        print_("creating vertex")
        dt = DateTimeNaiveTestVertex.create(test_val=datetime.datetime(2014, 1, 1))
        print_("getting vertex from vertex: %s" % dt)
        dt2 = DateTimeNaiveTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = DateTimeNaiveTestVertex.create(test_val=datetime.datetime(2014, 1, 1))
        print_("\ncreated vertex: %s" % dt)
        dt2 = DateTimeNaiveTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, datetime.datetime(2014, 1, 1))
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_datetime')
class TestVertexChoicesTestCase(BaseMogwaiTestCase):

    def test_good_choices_key_io(self):
        print_("creating vertex")
        dt = DateTimeNaiveTestChoicesVertex.create(test_val=datetime.datetime(2014, 1, 1))
        print_("validating input")
        self.assertEqual(dt.test_val, datetime.datetime(2014, 1, 1))
        print_("deleting vertex")
        dt.delete()

    @nottest
    def test_good_choices_value_io(self):
        # Known to be a bug, all keys and choices must be int | long | datetime
        print_("creating vertex")
        dt = DateTimeNaiveTestChoicesVertex.create(test_val='B')
        print_("validating input")
        self.assertEqual(dt.test_val, datetime.datetime(2014, 2, 1))
        print_("deleting vertex")
        dt.delete()

    def test_bad_choices_io(self):
        with self.assertRaises(ValidationError):
            print_("creating vertex")
            dt = DateTimeNaiveTestChoicesVertex.create(test_val=datetime.datetime(2014, 3, 1))
            print_("validating input")
            self.assertEqual(dt.test_val, 'C')
            print_("deleting vertex")
            dt.delete()



from __future__ import unicode_literals
from mogwai._compat import integer_types
from nose.plugins.attrib import attr
from mogwai.tests.base import BaseMogwaiTestCase
from mogwai.metrics.base import get_time, BaseMetricsReporter, MetricsRegistry
from mogwai.exceptions import MogwaiMetricsException


@attr('unit', 'metrics')
class BaseMetricReporterTestCase(BaseMogwaiTestCase):
    """
    Test Base Metric Reporter
    """

    def test_time_method(self):
        import time
        self.assertAlmostEqual(get_time(), time.time(), 0)

    def test_default_construction(self):
        mr = BaseMetricsReporter()
        self.assertIsInstance(mr.registry, (tuple, list))
        self.assertEqual(len(mr.registry), 1)
        self.assertIsInstance(mr.registry[0], MetricsRegistry)

    def test_single_reporter_construction(self):
        r = MetricsRegistry()
        mr = BaseMetricsReporter(registry=r)
        self.assertIsInstance(mr.registry, (tuple, list))
        self.assertEqual(len(mr.registry), 1)
        self.assertIsInstance(mr.registry[0], MetricsRegistry)
        self.assertEqual(mr.registry[0], r)

    def test_multiple_reporter_constuction(self):
        r = [MetricsRegistry(), MetricsRegistry()]
        mr = BaseMetricsReporter(registry=r)
        self.assertIsInstance(mr.registry, (tuple, list))
        self.assertEqual(len(mr.registry), 2)
        self.assertIsInstance(mr.registry[0], MetricsRegistry)
        self.assertEqual(mr.registry, r)

    def test_bad_construction(self):

        class NotMetricReporter(object):
            pass

        r = [NotMetricReporter()]
        with self.assertRaises(MogwaiMetricsException):
            mr = BaseMetricsReporter(registry=r)

    def test_start_stop_mechanism(self):
        mr = BaseMetricsReporter()
        mr.start()
        self.assertTrue(mr.task.running)
        mr.stop()
        self.assertFalse(mr.task.running)

    def test_metric_getter(self):
        mr = BaseMetricsReporter()
        mr.start()
        # increment a counter
        mr.registry[0].counter('test').inc()
        timestamp, metrics = mr.get_metrics()
        self.assertIsInstance(timestamp, integer_types)
        self.assertIsInstance(metrics, dict)
        self.assertEqual(len(metrics), 1)
        self.assertIn('test', metrics)
        self.assertIsInstance(metrics['test'], dict)
        self.assertEqual(metrics['test']['count'], 1)
        mr.stop()

    def test_metric_sender(self):
        mr = BaseMetricsReporter()
        mr.start()
        mr.registry[0].counter('test').inc()
        mr.send_metrics()


from __future__ import unicode_literals
from mogwai._compat import string_types, array_types
import logging
from re import compile
from rexpro.connection import RexProConnection
from rexpro.utils import get_rexpro
from rexpro.exceptions import RexProConnectionException, RexProScriptException

from mogwai.exceptions import MogwaiConnectionError, MogwaiQueryError
from mogwai.metrics.manager import MetricManager

logger = logging.getLogger(__name__)


SOCKET_TYPE = None
CONNECTION_TYPE = None
CONNECTION_POOL_TYPE = None
_connection_pool = None
_graph_name = None
metric_manager = MetricManager()
_loaded_models = []
__cached_spec = None


@metric_manager.time_calls
def execute_query(query, params={}, transaction=True, isolate=True, *args, **kwargs):
    """
    Execute a raw Gremlin query with the given parameters passed in.

    :param query: The Gremlin query to be executed
    :type query: str
    :param params: Parameters to the Gremlin query
    :type params: dict
    :param context: String context data to include with the query for stats logging
    :rtype: dict

    """
    global _connection_pool
    """ :type _connection_pool: RexProConnectionPool | None """

    if not _connection_pool:  # pragma: no cover
        raise MogwaiConnectionError('Must call mogwai.connection.setup before querying.')

    with _connection_pool.connection() as conn:
        try:
            response = conn.execute(query, params=params, isolate=isolate, transaction=transaction)

        except RexProConnectionException as ce:  # pragma: no cover
            _connection_pool.close_connection(conn, soft=True)
            raise MogwaiConnectionError("Connection Error during query - {}".format(ce))
        except RexProScriptException as se:  # pragma: no cover
            _connection_pool.close_connection(conn, soft=True)
            raise MogwaiQueryError("Error during query - {}".format(se))
        except:  # pragma: no cover
            _connection_pool.close_connection(conn, soft=True)
            raise

        logger.debug(response)
        return response


_host_re = compile(r'^((?P<user>.+?)(:(?P<password>.*?))?@)?(?P<host>.*?)(:(?P<port>\d+?))?(?P<graph_name>/.*?)?$')


def _parse_host(host, username, password, graph_name, graph_obj_name='g'):
        m = _host_re.match(host)
        d = m.groupdict() if m is not None else {}
        host = d.get('host', None) or '127.0.0.1'
        port = int(d.get('port', None) or 8184)
        username = d.get('username', None) or username
        password = d.get('password', None) or password
        graph_name = d.get('graph_name', None) or graph_name
        graph_obj_name = graph_obj_name or 'g'
        return {'host': host, 'port': port,
                'username': username, 'password': password,
                'graph_name': graph_name, 'graph_obj_name': graph_obj_name}


def setup(host, graph_name='graph', graph_obj_name='g', username='', password='',
          metric_reporters=None, pool_size=10, concurrency='sync'):
    """  Sets up the connection, and instantiates the models

    """
    global _connection_pool
    global SOCKET_TYPE, CONNECTION_TYPE, CONNECTION_POOL_TYPE
    global metric_manager

    if metric_reporters:  # pragma: no cover
        metric_manager.setup_reporters(metric_reporters)

    sock, conn, pool = get_rexpro(stype=concurrency)
    # store for reference
    SOCKET_TYPE = sock
    CONNECTION_TYPE = conn
    CONNECTION_POOL_TYPE = pool

    if isinstance(host, string_types):
        _connection_pool = pool(pool_size=pool_size,
                                **_parse_host(host, username, password, graph_name, graph_obj_name))

    else:  # pragma: no cover
        raise MogwaiConnectionError("Must Specify at least one host or list of hosts: host: {}, graph_name: {}".format(
            host, graph_name)
        )


def _add_model_to_space(model):
    global _loaded_models
    _loaded_models.append(model)


def generate_spec():
    """ Generates a titan index and type specification document based on loaded Vertex and Edge models """
    global _loaded_models, __cached_spec
    if __cached_spec:
        return __cached_spec
    from mogwai.models import Edge
    spec_list = []
    for model in _loaded_models:
        if not model.__abstract__ and hasattr(model, 'get_element_type'):

            makeType = 'makeLabel' if issubclass(model, Edge) else 'makeKey'
            element_type = 'Edge' if issubclass(model, Edge) else 'Vertex'

            spec = {'model': model.get_element_type(),
                    'element_type': element_type,
                    'makeType': makeType,
                    'properties': {}}
            for property in model._properties.values():
                if property.index:
                    # Only set this up for indexed properties

                    # Uniqueness constraint
                    uniqueness = ""
                    if property.unique and property.unique.lower() == 'in':
                        uniqueness = ".unique()"
                    elif property.unique and property.unique.lower() == 'out':
                        uniqueness = ".unidirected()"
                    elif property.unique and property.unique.lower() == 'both':
                        uniqueness = ".unique().single()"
                    if property.unique and property.unique.lower() == 'list':
                        uniqueness += ".list()"

                    # indexing extensions support
                    if not property.index_ext:
                        index_ext = ""
                    else:
                        index_ext = ".indexed(%s)" % property.index_ext

                    compiled_index = {"script": "g.{}(name).dataType({}.class).indexed({}{}.class){}.make(); g.commit()".format(
                        makeType,
                        property.data_type,
                        index_ext,
                        element_type,
                        uniqueness),
                                      "params": {'name': property.db_field_name},
                                      "transaction": False}
                    spec['properties'][property.db_field_name] = {
                        'data_type': property.data_type,
                        'index_ext': index_ext,
                        'uniqueness': uniqueness,
                        'compiled': compiled_index,
                    }

            spec_list.append(spec)
    __cached_spec = spec_list
    return spec_list


def sync_spec(filename, host, graph_name='graph', graph_obj_name='g', username='', password='', dry_run=False):  # pragma: no cover
    """
    Sync the given spec file to mogwai.

    :param filename: The filename of the spec file
    :type filename: str
    :param host: The host the be synced
    :type host: str
    :param graph_name: The name of the graph to be synced
    :type graph_name: str
    :param dry_run: Only prints generated Gremlin if True
    :type dry_run: boolean
    :returns: None

    """
    conn = RexProConnection(graph_name=graph_name, **_parse_host(host, username, password, graph_name, graph_obj_name))
    #Spec(filename).sync(conn, dry_run=dry_run)
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from mogwai._compat import print_
from nose.plugins.attrib import attr

from mogwai import connection
from mogwai.exceptions import MogwaiException
from mogwai.tests.base import BaseMogwaiTestCase, TestVertexModel, TestEdgeModel, TestVertexModelDouble

from mogwai import gremlin
from mogwai import models
from mogwai.models import Edge, Vertex
from mogwai.models.vertex import EnumVertexBaseMeta
from mogwai import properties
from mogwai._compat import with_metaclass
from mogwai.exceptions import MogwaiQueryError


class TestVertexModel2(Vertex):
    name = properties.String(default='test_text')
    test_val = properties.Integer()


class OtherTestModel(Vertex):
    name = properties.Text(default='test_text')
    test_val = properties.Integer(default=3)


class AliasedTestModel(Vertex):
    name = properties.Text()
    test_val = properties.Integer(db_field='how_many')


class OtherTestEdge(Edge):
    test_val = properties.Integer(default=1)


class YetAnotherTestEdge(Edge):
    test_val = properties.Integer(default=2)


@attr('unit', 'vertex_io')
class TestVertexIO(BaseMogwaiTestCase):

    def test_unicode_io(self):
        """
        Tests that unicode is saved and retrieved properly
        """
        tm1 = TestVertexModel.create(test_val=9, name=u'test2')
        tm2 = TestVertexModel.get(tm1._id)
        tm1.delete()

    def test_model_save_and_load(self):
        """
        Tests that models can be saved and retrieved
        """
        tm0 = TestVertexModel.create(test_val=8, name='123456789')
        tm1 = TestVertexModel.create(test_val=9, name='456789')
        tms = TestVertexModel.all([tm0.id, tm1.id])

        self.assertEqual(len(tms), 2)

        for pname in tm0._properties.keys():
            self.assertEquals(getattr(tm0, pname), getattr(tms[0], pname))

        tms = TestVertexModel.all([tm1.id, tm0.id])
        self.assertEqual(tms[0].id, tm1.id)
        self.assertEqual(tms[1].id, tm0.id)

        tm0.delete()
        tm1.delete()

    def test_model_updating_works_properly(self):
        """
        Tests that subsequent saves after initial model creation work
        """
        tm = TestVertexModel.create(test_val=8, name='123456789')

        tm.test_val = 100
        tm.save()

        tm.test_val = 80
        tm.save()

        tm.test_val = 60
        tm.save()

        tm.test_val = 40
        tm.save()

        tm.test_val = 20
        tm.save()

        tm2 = TestVertexModel.get(tm.id)
        self.assertEquals(tm.test_val, tm2.test_val)
        tm.delete()

    def test_model_deleting_works_properly(self):
        """
        Tests that an instance's delete method deletes the instance
        """
        tm = TestVertexModel.create(test_val=8, name='123456789')
        vid = tm.id
        tm.delete()
        with self.assertRaises(TestVertexModel.DoesNotExist):
            tm2 = TestVertexModel.get(vid)

    def test_reload(self):
        """ Tests that and instance's reload method does an inplace update of the instance """
        tm0 = TestVertexModel.create(test_val=8, name='123456789')
        tm1 = TestVertexModel.get(tm0.id)
        tm1.test_val = 7
        tm1.save()

        tm0.reload()
        self.assertEqual(tm0.test_val, 7)
        tm0.delete()

    def test_reload_on_aliased_field(self):
        """ Tests that reload works with aliased fields """
        tm0 = AliasedTestModel.create(test_val=8, name='123456789')
        tm1 = AliasedTestModel.get(tm0.id)
        tm1.test_val = 7
        tm1.save()

        tm0.reload()
        self.assertEqual(tm0.test_val, 7)
        tm1.delete()

    def test_all_method(self):
        with self.assertRaises(MogwaiQueryError):
            TestVertexModel.all(1)

    def test_all_method_invalid_length(self):
        v1 = TestVertexModel.create()
        v2 = TestVertexModel.create()
        from mogwai.exceptions import MogwaiQueryError
        with self.assertRaises(MogwaiQueryError):
            TestVertexModel.all([v1.id, v2.id, 'invalid'])
        v1.delete()
        v2.delete()

    def test_find_by_value_method(self):
        v1 = TestVertexModel.create(name='v1', test_val=-99)
        v2 = TestVertexModel.create(name='v2', test_val=-99)
        v3 = TestVertexModelDouble.create(name='v3', test_val=-100.0)

        self.assertEqual(len(TestVertexModel.find_by_value('name', 'v1')), 1)
        self.assertEqual(len(TestVertexModel.find_by_value('test_val', -99)), 2)
        self.assertEqual(len(TestVertexModel.find_by_value('name', 'bar')), 0)

        self.assertEqual(TestVertexModel.find_by_value('name', 'v1')[0], v1)

        self.assertEqual(len(TestVertexModelDouble.find_by_value('test_val', -100.0)), 1)
        v1.delete()
        v2.delete()
        v3.delete()

    def test_get_by_id(self):
        v1 = TestVertexModel.create()
        results = TestVertexModel.get(v1.id)
        self.assertIsInstance(results, TestVertexModel)
        self.assertEqual(results, v1)

        with self.assertRaises(TestEdgeModel.DoesNotExist):
            results = TestVertexModel.get(None)

        with self.assertRaises(TestEdgeModel.DoesNotExist):
            results = TestVertexModel.get('nonexistant')

        v2 = TestVertexModel2.create(test_val=0)
        with self.assertRaises(TestVertexModel.WrongElementType):
            results = TestVertexModel.get(v2.id)

        v2.delete()
        v1.delete()

    @attr('vertex_delete_methods')
    def test_delete_methods(self):
        v1 = TestVertexModel.create()
        v2 = TestVertexModel.create()

        # delete_outE
        e1 = TestEdgeModel.create(v1, v2)
        v1.delete_outE(TestEdgeModel)

        # delete_inE
        e1 = TestEdgeModel.create(v1, v2)
        v2.delete_inE(TestEdgeModel)

        # delete_inV
        e1 = TestEdgeModel.create(v1, v2)
        v1.delete_inV(TestEdgeModel)

        # delete_outV
        v2 = TestVertexModel.create()
        e1 = TestEdgeModel.create(v1, v2)
        v2.delete_outV(TestEdgeModel)

        v2.delete()


class DeserializationTestModel(Vertex):
    count = properties.Integer()
    text = properties.Text()

    gremlin_path = 'deserialize.groovy'

    get_map = gremlin.GremlinValue()
    get_list = gremlin.GremlinMethod()


@attr('unit', 'vertex_io')
class TestNestedDeserialization(BaseMogwaiTestCase):
    """
    Tests that vertices are properly deserialized when nested in map and list data structures
    """

    def test_map_deserialization(self):
        """
        Tests that elements nested in maps are properly deserialized
        """

        original = DeserializationTestModel.create(count=5, text='happy')
        nested = original.get_map()

        self.assertIsInstance(nested, dict)
        self.assertEqual(nested['vertex'], original)
        self.assertEqual(nested['number'], 5)
        original.delete()

    def test_list_deserialization(self):
        """
        Tests that elements nested in lists are properly deserialized
        """

        original = DeserializationTestModel.create(count=5, text='happy')
        nested = original.get_list()

        self.assertIsInstance(nested, list)
        self.assertEqual(nested[0], None)
        self.assertEqual(nested[1], 0)
        self.assertEqual(nested[2], 1)

        self.assertIsInstance(nested[3], list)
        self.assertEqual(nested[3][0], 2)
        self.assertEqual(nested[3][1], original)
        self.assertEqual(nested[3][2], 3)

        self.assertEqual(nested[4], 5)
        original.delete()


class TestEnumVertexModel(with_metaclass(EnumVertexBaseMeta, Vertex)):
    __enum_id_only__ = False
    name = properties.String(default='test text')
    test_val = properties.Integer(default=0)

    def enum_generator(self):
        return '%s_%s' % (self.name.replace(' ', '_').upper(), self.test_val)


class TestEnumVertexModel2(with_metaclass(EnumVertexBaseMeta, Vertex)):
    name = properties.String(default='test text')
    test_val = properties.Integer(default=0)


@attr('unit', 'vertex_io', 'vertex_enum')
class TestVertexEnumModel(BaseMogwaiTestCase):

    def test_default_enum_handling(self):
        """ Default enum handling test

        Works by grabbing the `name` property and substituting ' ' for '_' and capitalizing
        The enum returns the Vertex ID
        """
        tv = TestEnumVertexModel2.create()
        self.assertEqual(TestEnumVertexModel2.TEST_TEXT, tv.id)
        tv.delete()

    def test_custom_enum_handling(self):
        """ Custom enum handling test

        Works utilizing the VertexModel's `enum_generator` instance method to generate the ENUMs
        The enum returns the actual Vertex instance because the `__enum_id_only__` was set to False
        which stores the entire Vertex instead of just the ID.
        """
        tv = TestEnumVertexModel.create()
        self.assertEqual(TestEnumVertexModel.TEST_TEXT_0, tv)
        tv.delete()

    def test_attempt_load_fail(self):
        with self.assertRaises(AttributeError):
            TestEnumVertexModel.WRONG


@attr('unit', 'vertex_io')
class TestUpdateMethod(BaseMogwaiTestCase):
    def test_success_case(self):
        """ Tests that the update method works as expected """
        tm = TestVertexModel.create(test_val=8, name='123456789')
        tm2 = tm.update(test_val=9)

        tm3 = TestVertexModel.get(tm.id)
        self.assertEqual(tm2.test_val, 9)
        self.assertEqual(tm3.test_val, 9)
        tm.delete()

    def test_unknown_names_raise_exception(self):
        """ Tests that passing in names for columns that don't exist raises an exception """
        tm = TestVertexModel.create(test_val=8, text='123456789')
        with self.assertRaises(TypeError):
            tm.update(jon='beard')
        tm.delete()


@attr('unit', 'vertex_io')
class TestVertexTraversal(BaseMogwaiTestCase):
    def setUp(self):
        super(TestVertexTraversal, self).setUp()
        self.v1 = TestVertexModel.create(test_val=1, name='Test1')
        self.v2 = TestVertexModel.create(test_val=2, name='Test2')
        self.v3 = OtherTestModel.create(test_val=3, name='Test3')
        self.v4 = OtherTestModel.create(test_val=3, name='Test3')

    def tearDown(self):
        self.v1.delete()
        self.v2.delete()
        self.v3.delete()
        self.v4.delete()

    def test_outgoing_vertex_traversal(self):
        """Test that outgoing vertex traversals work."""
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=12)
        e2 = TestEdgeModel.create(self.v1, self.v3, test_val=13)
        e3 = TestEdgeModel.create(self.v2, self.v3, test_val=14)

        results = self.v1.outV(TestEdgeModel)
        self.assertEqual(len(results), 2)
        self.assertIn(self.v2, results)
        self.assertIn(self.v3, results)

        results = self.v1.outV(TestEdgeModel, types=[OtherTestModel])
        self.assertEqual(len(results), 1)
        self.assertIn(self.v3, results)

        e1.delete()
        e2.delete()
        e3.delete()

    def test_incoming_vertex_traversal(self):
        """Test that incoming vertex traversals work."""
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=12)
        e2 = TestEdgeModel.create(self.v1, self.v3, test_val=13)
        e3 = TestEdgeModel.create(self.v2, self.v3, test_val=14)

        results = self.v2.inV(TestEdgeModel)
        self.assertEqual(len(results), 1)
        self.assertIn(self.v1, results)

        results = self.v2.inV(TestEdgeModel, types=[OtherTestModel])
        self.assertEqual(len(results), 0)

        e1.delete()
        e2.delete()
        e3.delete()

    def test_outgoing_edge_traversals(self):
        """Test that outgoing edge traversals work."""
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=12)
        e2 = TestEdgeModel.create(self.v1, self.v3, test_val=13)
        e3 = OtherTestEdge.create(self.v2, self.v3, test_val=14)

        results = self.v2.outE()
        self.assertEqual(len(results), 1)
        self.assertIn(e3, results)

        results = self.v2.outE(types=[TestEdgeModel])
        self.assertEqual(len(results), 0)

        e1.delete()
        e2.delete()
        e3.delete()

    def test_incoming_edge_traversals(self):
        """Test that incoming edge traversals work."""
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=12)
        e2 = TestEdgeModel.create(self.v1, self.v3, test_val=13)
        e3 = OtherTestEdge.create(self.v2, self.v3, test_val=14)

        results = self.v2.inE()
        self.assertEqual(len(results), 1)
        self.assertIn(e1, results)

        results = self.v2.inE(types=[OtherTestEdge])
        self.assertEqual(len(results), 0)

        e1.delete()
        e2.delete()
        e3.delete()

    def test_multiple_label_traversals(self):
        """ Tests that using multiple edges for traversals works """
        e1 = TestEdgeModel.create(self.v1, self.v2)
        e2 = OtherTestEdge.create(self.v1, self.v3)
        e3 = YetAnotherTestEdge.create(self.v1, self.v4)

        self.assertEqual(len(self.v1.outV()), 3)

        self.assertEqual(len(self.v1.outV(TestEdgeModel)), 1)
        self.assertEqual(len(self.v1.outV(OtherTestEdge)), 1)
        self.assertEqual(len(self.v1.outV(YetAnotherTestEdge)), 1)

        out = self.v1.outV(TestEdgeModel, OtherTestEdge)
        self.assertEqual(len(out), 2)
        self.assertIn(self.v2.id, [v.id for v in out])
        self.assertIn(self.v3.id, [v.id for v in out])

        out = self.v1.outV(OtherTestEdge, YetAnotherTestEdge)
        self.assertEqual(len(out), 2)
        self.assertIn(self.v3.id, [v.id for v in out])
        self.assertIn(self.v4.id, [v.id for v in out])

        e1.delete()
        e2.delete()
        e3.delete()

    def test_multiple_edge_traversal_with_type_filtering(self):
        """ Tests that using multiple edges for traversals works """
        v = TestVertexModel.create(test_val=1, name='Test1')

        v1 = TestVertexModel.create()
        e1 = TestEdgeModel.create(v, v1)

        v2 = TestVertexModel.create()
        e2 = OtherTestEdge.create(v, v2)

        v3 = TestVertexModel.create()
        e3 = YetAnotherTestEdge.create(v, v3)

        v4 = OtherTestModel.create()
        e4 = TestEdgeModel.create(v, v4)

        v5 = OtherTestModel.create()
        e5 = OtherTestEdge.create(v, v5)

        v6 = OtherTestModel.create()
        e6 = YetAnotherTestEdge.create(v, v6)

        self.assertEqual(len(v.outV()), 6)

        self.assertEqual(len(v.outV(TestEdgeModel, OtherTestEdge)), 4)
        self.assertEqual(len(v.outV(TestEdgeModel, OtherTestEdge, types=[TestVertexModel])), 2)

        e1.delete()
        e2.delete()
        e3.delete()
        e4.delete()
        e5.delete()
        e6.delete()
        v.delete()
        v1.delete()
        v2.delete()
        v3.delete()
        v4.delete()
        v5.delete()
        v6.delete()

    def test_edge_instance_traversal_types(self):
        """ Test traversals with edge instances work properly """
        te = TestEdgeModel.create(self.v1, self.v2)
        ote = OtherTestEdge.create(self.v1, self.v3)
        yate = YetAnotherTestEdge.create(self.v1, self.v4)

        out = self.v1.outV(te, ote)
        self.assertEqual(len(out), 2)
        self.assertIn(self.v2.id, [v.id for v in out])
        self.assertIn(self.v3.id, [v.id for v in out])

        out = self.v1.outV(ote, yate)
        self.assertEqual(len(out), 2)
        self.assertIn(self.v3.id, [v.id for v in out])
        self.assertIn(self.v4.id, [v.id for v in out])

        te.delete()
        ote.delete()
        yate.delete()

    def test_edge_label_string_traversal_types(self):
        """ Test traversals with edge instances work properly """
        e1 = TestEdgeModel.create(self.v1, self.v2)
        e2 = OtherTestEdge.create(self.v1, self.v3)
        e3 = YetAnotherTestEdge.create(self.v1, self.v4)

        out = self.v1.outV(TestEdgeModel.get_label(), OtherTestEdge.get_label())
        self.assertEqual(len(out), 2)
        self.assertIn(self.v2.id, [v.id for v in out])
        self.assertIn(self.v3.id, [v.id for v in out])

        out = self.v1.outV(OtherTestEdge.get_label(), YetAnotherTestEdge.get_label())
        self.assertEqual(len(out), 2)
        self.assertIn(self.v3.id, [v.id for v in out])
        self.assertIn(self.v4.id, [v.id for v in out])

        e1.delete()
        e2.delete()
        e3.delete()

    def test_unknown_edge_traversal_filter_type_fails(self):
        """
        Tests an exception is raised if a traversal filter is
        used that's not an edge class, instance or label string fails
        """
        e1 = TestEdgeModel.create(self.v1, self.v2)
        e2 = OtherTestEdge.create(self.v1, self.v3)
        e3 = YetAnotherTestEdge.create(self.v1, self.v4)

        with self.assertRaises(MogwaiException):
            out = self.v1.outV(5)

        with self.assertRaises(MogwaiException):
            out = self.v1.outV(True)

        e1.delete()
        e2.delete()
        e3.delete()

    @attr('manual_properties')
    def test_manual_properties(self):
        v = TestVertexModel.create(test_val=1, name='Test 7', some_property=32)

        print_("Got Results: {}".format(v))
        print_("Result dict: {}".format(v.as_dict()))
        print_("\tResult properties: {}".format(v._properties.keys()))
        print_("\tResult Manaul: {}".format(v._manual_values.items()))

        self.assertEqual(v['some_property'], 32)
        self.assertIn('some_property', v)  # This also tests __contains__
        self.assertIn('test_val', v)  # This also tests __contains__

        # test len(Element()), should return len(element._properties) + len(element._manual_values)
        self.assertEqual(len(v), 3)

        # test __iter__
        prop_keys = [key for key in v]
        self.assertEqual(len(prop_keys), 3)
        for prop_key in v:
            self.assertIn(prop_key, ['test_val', 'name', 'some_property'])

        # test keys()
        self.assertEqual(len(prop_keys), len(v.keys()))
        for prop_key in v.keys():
            self.assertIn(prop_key, prop_keys)

        # test values()
        prop_values = v.values()
        self.assertEqual(len(prop_values), 3)
        for prop_value in prop_values:
            self.assertIn(prop_value, [1, 'Test 7', 32])

        # test items()
        prop_items = v.items()
        self.assertEqual(len(prop_items), 3)
        for prop_key, prop_value in prop_items:
            self.assertIn(prop_key, ['test_val', 'name', 'some_property'])
            self.assertIn(prop_value, [1, 'Test 7', 32])

        # test change
        v['some_property'] = 42
        self.assertEqual(v['some_property'], 42)
        v.save()
        self.assertEqual(v['some_property'], 42)

        # test delete
        del v['some_property']
        # This should still exist, so the property can be removed from the database,
        # but should raise an AttributeError if attempted to access normally
        self.assertIn('some_property', v)
        self.assertIsNone(v._manual_values.get('some_property'))
        with self.assertRaises(AttributeError):
            value = v['some_property']
            print_("Got value: {}".format(value))

        v.delete()



from mogwai._compat import print_
from .base import BaseMetricsReporter
from mogwai.exceptions import MogwaiMetricsException
from pyformance import call_too_long
import time
from functools import wraps
import logging


logger = logging.getLogger(__name__)


class TimerContext(object):
    """ Timer Context Manager customization to compensate for multiple Metric Registries """
    def __init__(self, timers, clock=time, *args, **kwargs):
        super(TimerContext, self).__init__()
        self.clock = clock
        self.timers = timers
        self.start_time = self.clock.time()
        self.kwargs = kwargs
        self.args = args

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed = self.clock.time() - self.start_time
        for timer in self.timers:
            """ @type timer : pyformance.meters.timer.Timer """
            timer._update(elapsed)
            if timer.threshold and timer.threshold < elapsed:  # pragma: no cover
                # this is a future feature in pyformance and will need to be tested
                call_too_long.send(timer, elapsed=elapsed, *self.args, **self.kwargs)
        if exc_type is not None:
            return False
        return elapsed


class MetricManager(object):
    """ Manages your Metric Agents

    Properly inject mogwai metrics, while still allowing custom app metrics
    """

    def __init__(self):
        self.metric_reporters = []  #BaseMetricsReporter()]

    def setup_reporters(self, metric_reporters):
        """ Setup the Metric Reporter(s) for the MetricManager """
        if not metric_reporters:
            raise MogwaiMetricsException("No Metric Reporters provided!")
        if not isinstance(metric_reporters, (list, tuple)):
            metric_reporters = [metric_reporters, ]

        for mr in metric_reporters:
            if not isinstance(mr, BaseMetricsReporter):
                raise MogwaiMetricsException("{} Not derived from Mogwai BaseMetricsReporter".format(mr))

        self.metric_reporters = metric_reporters

    def start(self):
        """ Start the Metric Reporter """
        for mr in self.metric_reporters:
            mr.start()

    def stop(self):
        """ Stop the Metric Reporter """
        for mr in self.metric_reporters:
            mr.stop()

    def timers(self, key=None):
        """ Yield Metric Reporter Timers

        :param key: The metric key to use
        :type key: basestring
        """
        if not key:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.timer
        else:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.timer(key)

    def histograms(self, key=None):
        """ Yield Metric Reporter Histograms

        :param key: The metric key to use
        :type key: basestring
        """
        if not key:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.histogram
        else:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.histogram(key)

    def meters(self, key=None):
        """ Yield Metric Reporter Meters

        :param key: The metric key to use
        :type key: basestring
        """
        if not key:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.meter
        else:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.meter(key)

    def counters(self, key=None):
        """ Yield Metric Reporter Counters

        :param key: The metric key to use
        :type key: basestring
        """
        if not key:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.counter
        else:
            for mr in self.metric_reporters:
                for reg in mr.registry:
                    yield reg.counter(key)

    def time_calls(self, fn):
        """
        Decorator to time the execution of the function.

        :param fn: the function to be decorated
        :type fn: C{func}
        :return: the decorated function
        :rtype: C{func}
        """
        @wraps(fn)
        def wrapper(*args, **kwargs):
            fn_name = "{}.timer".format(fn.__name__)
            if 'context' in kwargs:
                context_name = "{}.timer".format(kwargs.get('context'))
                del kwargs['context']
                timers = [reg.timer(context_name) for mr in self.metric_reporters for reg in mr.registry]
            else:
                timers = []
            timers += [reg.timer(fn_name) for mr in self.metric_reporters for reg in mr.registry]
            with TimerContext(timers):
                try:
                    return fn(*args, **kwargs)
                except:
                    for mr in self.metric_reporters:
                        for reg in mr.registry:
                            reg.meter("mogwai.error.meter").mark()
                            reg.counter("mogwai.error").inc()
                    raise
        return wrapper

    def hist_calls(self, fn):
        """
        Decorator to check the distribution of return values of a function.

        :param fn: the function to be decorated
        :type fn: C{func}
        :return: the decorated function
        :rtype: C{func}
        """
        @wraps(fn)
        def wrapper(*args, **kwargs):
            fn_name = "{}.calls".format(fn.__name__)
            context_name = None
            if 'context' in kwargs:
                context_name = "{}.hist".format(kwargs.get('context'))
                del kwargs['context']

            try:
                rtn = fn(*args, **kwargs)
                if isinstance(rtn, (int, float)):
                    for mr in self.metric_reporters:
                        for reg in mr.registry:
                            reg.histogram(fn_name).add(rtn)
                            if context_name:
                                reg.histogram(context_name).add(rtn)
                return rtn
            except:
                for mr in self.metric_reporters:
                    for reg in mr.registry:
                        reg.meter("mogwai.error.meter").mark()
                        reg.counter("mogwai.error").inc()
                raise
        return wrapper

    def meter_calls(self, fn):
        """
        Decorator to the rate at which a function is called.

        :param fn: the function to be decorated
        :type fn: C{func}
        :return: the decorated function
        :rtype: C{func}
        """
        @wraps(fn)
        def wrapper(*args, **kwargs):
            fn_name = "{}.calls".format(fn.__name__)
            context_name = None
            if 'context' in kwargs:
                context_name = "{}.meter".format(kwargs.get('context'))
                del kwargs['context']

            for mr in self.metric_reporters:
                for reg in mr.registry:
                    reg.meter(fn_name).mark()
                    if context_name:
                        reg.meter(context_name).mark()
            try:
                return fn(*args, **kwargs)
            except:
                for mr in self.metric_reporters:
                    for reg in mr.registry:
                        reg.meter("mogwai.error.meter").mark()
                        reg.counter("mogwai.error").inc()
                raise
        return wrapper

    def count_calls(self, fn):
        """
        Decorator to track the number of times a function is called.

        :param fn: the function to be decorated
        :type fn: C{func}
        :return: the decorated function
        :rtype: C{func}
        """
        @wraps(fn)
        def wrapper(*args, **kwargs):
            fn_name = "{}.calls".format(fn.__name__)
            context_name = None
            if 'context' in kwargs:
                context_name = "{}.counter".format(kwargs.get('context'))
                del kwargs['context']

            for mr in self.metric_reporters:
                for reg in mr.registry:
                    reg.counter(fn_name).inc()
                    if context_name:
                        reg.counter(context_name).inc()
            #print_("Trying function: %s" % fn_name)
            try:
                return fn(*args, **kwargs)
            except:
                for mr in self.metric_reporters:
                    for reg in mr.registry:
                        reg.meter("mogwai.error.meter").mark()
                        reg.counter("mogwai.error").inc()
                raise
        return wrapper

__all__ = ['MetricManager', 'MogwaiMetricsException']



from __future__ import unicode_literals
from unittest import TestCase
from nose.tools import nottest
from mogwai.connection import setup, sync_spec
from mogwai.models import Vertex, Edge
from mogwai.properties import Double, Integer, String
import os

_val = 0


def counter():
    global _val
    _val += 1
    return _val


class TestVertexModel(Vertex):
    element_type = 'test_vertex_model'

    name = String(default='test_vertex')
    test_val = Integer(default=counter)


class TestVertexModelDouble(Vertex):
    element_type = 'test_vertex_model_double'

    name = String(default='test_vertex_double')
    test_val = Double(default=0.0)


class TestEdgeModel(Edge):
    label = 'test_edge_model'

    name = String(default='test_edge')
    test_val = Integer(default=counter)


class TestEdgeModel2(Edge):
    label = 'test_edge_model_2'

    name = String(default='test_edge')
    test_val = Integer(default=counter)


class TestEdgeModelDouble(Edge):
    label = 'test_edge_model_double'

    name = String(default='test_edge_double')
    test_val = Double(default=0.0)


@nottest
def testcase_docstring_sub(*sub):
    """ If you wanted to lazy load something into a docstring on a test.

    Do not use this normally, it's bad practice.
    """
    def decorated(obj):
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj
    return decorated


class BaseMogwaiTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        super(BaseMogwaiTestCase, cls).setUpClass()
        #sync_spec(filename='test.spec', host='192.168.133.12', graph_name='graph')
        setup(os.getenv('TITAN_REXPRO_URL', 'localhost'), graph_name='graph')

    def assertHasAttr(self, obj, attr):
        self.assertTrue(hasattr(obj, attr), "%s doesn't have attribute: %s" % (obj, attr))

    def assertNotHasAttr(self, obj, attr):
        self.assertFalse(hasattr(obj, attr), "%s shouldn't have attribute: %s" % (obj, attr))

    def assertAttrEqual(self, obj, attr, value):
        self.assertHasAttr(obj, attr)
        self.assertEqual(getattr(obj, attr), value)

    def assertAttrNotEqual(self, obj, attr, value):
        self.assertHasAttr(obj, attr)
        self.assertNotEqual(getattr(obj, attr), value)

    def assertNotRaise(self, callableObj, *args, **kwargs):
        try:
            callableObj(*args, **kwargs)
        except Exception as e:
            raise AssertionError("Shouldn't raise and exception: {}".format(e))

    def assertAnyRaise(self, callableObj, *args, **kwargs):
        try:
            callableObj(*args, **kwargs)
        except:
            return
        raise AssertionError("Should raise an exception")

    def assertIsSubclass(self, C, B):
        if issubclass(C, B):
            return
        else:
            raise AssertionError("{} is Not a Subclass of {}".format(B, C))

    def assertDictContainsKey(self, obj, key):
        if key in obj:
            return
        else:
            raise AssertionError("{} is not in dict: {}".format(key, obj))

    def assertDictContainsKeyWithValueType(self, obj, key, B):
        if key in obj:
            if isinstance(obj[key], B):
                return
            else:
                raise AssertionError("dict[{}]={} is not of type: {}".format(key, obj[key], B))
        else:
            raise AssertionError("{} is not in dict: {}".format(key, obj))

    def assertIsVertex(self, obj):
        self.assertIsSubclass(obj, Vertex)

    def assertIsEdge(self, obj):
        self.assertIsSubclass(obj, Edge)



## TODO: Reference: Tooling from werkzeug tools - citation needed
from __future__ import unicode_literals
import sys
from mogwai._compat import reraise, string_types, PY2
from factory import base
import factory as factory


class ImportStringError(ImportError):
    """Provides information about a failed :func:`import_string` attempt."""

    # String in dotted notation that failed to be imported.
    import_name = None
    # Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            'import_string() failed for %r. Possible reasons are:\n\n'
            '- missing __init__.py in a package;\n'
            '- package or module path not included in sys.path;\n'
            '- duplicated package or module name taking precedence in '
            'sys.path;\n'
            '- missing module, class, function or variable;\n\n'
            'Debugged import:\n\n%s\n\n'
            'Original exception:\n\n%s: %s')

        name = ''
        tracked = []
        for part in import_name.replace(':', '.').split('.'):
            name += (name and '.') + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, '__file__', None)))
            else:
                track = ['- %r found in %r.' % (n, i) for n, i in tracked]
                track.append('- %r not found.' % name)
                msg = msg % (import_name, '\n'.join(track),
                             exception.__class__.__name__, str(exception))
                break

        ImportError.__init__(self, msg)

    def __repr__(self):  # pragma: no cover
        return '<%s(%r, %r)>' % (self.__class__.__name__, self.import_name,
                                 self.exception)


def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    assert isinstance(import_name, string_types)
    # force the import name to automatically convert to strings
    import_name = str(import_name)
    try:
        if ':' in import_name:
            module, obj = import_name.split(':', 1)
        elif '.' in import_name:
            module, obj = import_name.rsplit('.', 1)
        else:
            return __import__(import_name)
        # __import__ is not able to handle unicode setup_strings in the fromlist
        # if the module is a package
        if PY2 and isinstance(obj, unicode):  # pragma: no cover
            obj = obj.encode('utf-8')
        try:
            return getattr(__import__(module, None, None, [obj]), obj)
        except (ImportError, AttributeError):
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            modname = module + '.' + obj
            __import__(modname)
            return sys.modules[modname]
    except ImportError as e:
        if not silent:
            reraise(
                ImportStringError,
                ImportStringError(import_name, e),
                sys.exc_info()[2])


class _Missing(object):

    def __repr__(self):  # pragma: no cover
        return 'no value'

    def __reduce__(self):  # pragma: no cover
        return '_missing'

_missing = _Missing()


class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """

    # implementation detail: this property is implemented as non-data
    # descriptor.  non-data descriptors are only invoked if there is
    # no entry with the same name in the instance's __dict__.
    # this allows us to completely get rid of the access function call
    # overhead.  If one choses to invoke __get__ by hand the property
    # will still work as expected because the lookup logic is replicated
    # in __get__ for manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class LazyImportClass(object):
    """ Lazily load and return a class """

    def __init__(self, import_name):
        try:
            self.__module__, self.__name__ = import_name.split('.', 1)
        except ValueError:  # pragma: no cover
            self.__module__ = ''
            self.__name__ = import_name
        self.__module__ = self.__module__ .lstrip('.')
        self.import_name = import_name.lstrip('.')
        #self.__setup_instantiated_vertex = getattr(self.klass, '__setup_instantiated_vertex')

    @cached_property
    def klass(self):
        """ Imports the class and caches """
        return import_string(self.import_name)

    def __call__(self, *args, **kwargs):
        return self.klass(*args, **kwargs)

    def __setup_instantiated_vertex(self, *args, **kwargs):  # pragma: no cover
        return getattr(self.klass, '__setup_instantiated_vertex')(*args, **kwargs)


class Factory(base.Factory):
    """Factory for Mogwai models. """

    ABSTRACT_FACTORY = True

    _associated_model = None

    @classmethod
    def _load_target_class(cls, class_definition):
        """ So we can support potential circular import problems, by using normal import_string import specification.
        """
        associated_class = super(Factory, cls)._load_target_class(class_definition)

        if isinstance(associated_class, string_types) and '.' in associated_class:
            if cls._associated_model is None:
                cls._associated_model = import_string(associated_class)
                cls._associated_model.FACTORY_CLASS = cls
            return cls._associated_model

        if associated_class and associated_class.FACTORY_CLASS is None:
            associated_class.FACTORY_CLASS = cls

        return associated_class

    @classmethod
    def _create(cls, target_class, *args, **kwargs):
        """Create an instance of the model, and save it to the database."""
        from mogwai.models import Edge
        if issubclass(target_class.__class__, Edge):
            assert ('outV' in kwargs and 'inV' in kwargs), "Edges require in and out Vertices"
            obj = target_class.create(outV=kwargs.get('outV'), inV=kwargs.get('inV'), *args, **kwargs)
        else:
            obj = target_class.create(*args, **kwargs)
        return obj



from __future__ import unicode_literals
import datetime
from pytz import utc
from uuid import uuid4
from nose.plugins.attrib import attr

from mogwai.exceptions import MogwaiGremlinException, MogwaiException
from mogwai.tests.base import BaseMogwaiTestCase

from mogwai.models import Vertex
from mogwai import properties
from mogwai import gremlin
from mogwai.gremlin.table import Table, Row
from mogwai._compat import print_
from copy import deepcopy
from collections import OrderedDict


class GroovyTestModel2(Vertex):

    element_type = 'groovy_test_model2'

    text = properties.Text()

    get_table_of_models = gremlin.GremlinTable(path='groovy_test_model.groovy', classmethod=True, defaults={'element_type': 'groovy_test_model2'})


@attr('unit', 'gremlin', 'gremlin_table')
class TestGremlinTable(BaseMogwaiTestCase):

    def test_method_loads_and_works(self):
        elements = []
        for i in range(10):
            elements.append(GroovyTestModel2.create(text='test{}'.format(i)))

        table = GroovyTestModel2.get_table_of_models()

        self.assertIsInstance(table, Table)
        self.assertEqual(10, len(table))
        for row in table:
            self.assertIsInstance(row, Row)
            self.assertTrue(row.text.startswith('test'))

        for element in elements:
            element.delete()


@attr('unit', 'gremlin', 'gremlin_table', 'gremlin_table_table')
class TestTableContainer(BaseMogwaiTestCase):

    def test_gremlin_table_row(self):
        d = OrderedDict({'test': 1, 'a': 'b', 'c': 3})
        r = Row(deepcopy(d))

        print_(r)

        self.assertDictEqual(d, r._Row__data)
        self.assertEqual(len(d), len(r))

        # test tuple-like access
        self.assertIn(r[0], d.values())
        self.assertListEqual(list(d.values()), list(r[:]))
        with self.assertRaises(MogwaiException):
            r[0] = 'test'

        with self.assertRaises(MogwaiException):
            r[:] = (1, 2, 3)

        with self.assertRaises(MogwaiException):
            del r[0]

        with self.assertRaises(MogwaiException):
            del r[0:2]

        # test dict-like access
        self.assertListEqual(list(d.values()), list(r.values()))
        self.assertListEqual(list(d.keys()), list(r.keys()))
        self.assertListEqual(list(d.items()), list(r.items()))
        self.assertEqual(d['test'], r['test'])
        self.assertEqual(d['a'], r['a'])
        self.assertEqual(d['c'], r['c'])
        with self.assertRaises(MogwaiException):
            r['test'] = 0

        with self.assertRaises(MogwaiException):
            del r['test']

        # test . notation access
        self.assertEqual(d['test'], r.test)
        self.assertEqual(d['a'], r.a)
        self.assertEqual(d['c'], r.c)
        with self.assertRaises(MogwaiException):
            r.test = 0
        with self.assertRaises(MogwaiException):
            del r.test

        # test iteration
        self.assertEqual(len(d), len([v for v in r]))
        self.assertListEqual(list(d.values()), [v for v in r])

        # test next
        self.assertIn(r.next(), d.values())
        self.assertIn(r.next(), d.values())
        self.assertIn(r.next(), d.values())

        # test equality
        r2 = Row(deepcopy(d))
        self.assertEqual(r, r2)

    def test_gremlin_table_table(self):
        data = [
            OrderedDict({'test': 1, 'a': 'b', 'c': 3}),
            OrderedDict({'test': 2, 'a': 'd', 'c': 5})
        ]
        t = Table(deepcopy(data))

        self.assertListEqual(data, t._Table__gremlin_result)

        # test iterator
        self.assertListEqual(data, [r._Row__data for r in t])

        # test tuple-like access
        self.assertIsInstance(t[0], Row)
        self.assertDictEqual(data[0], t[0]._Row__data)
        self.assertEqual(data[0]['test'], t[0].test)
        with self.assertRaises(MogwaiException):
            t[0] = 'test'
        with self.assertRaises(MogwaiException):
            del t[0]

        self.assertEqual(t[0], t.next())
        self.assertEqual(t[1], t.next())



from __future__ import unicode_literals
from mogwai._compat import print_
import os
from nose.plugins.attrib import attr
from mogwai.tests.base import BaseMogwaiTestCase
from mogwai.gremlin.groovy import parse, GroovyImportParser, GroovyFunctionParser,\
    GroovyImport, GroovyFunction, GroovyFileDef


@attr('unit', 'groovy')
class GroovyScannerTest(BaseMogwaiTestCase):
    """
    Test Groovy language scanner
    """

    @classmethod
    def setUpClass(cls):
        cls.groovy_file = os.path.join(os.path.dirname(__file__), 'groovy_test_model.groovy')
        with open(cls.groovy_file, 'r') as f:
            cls.groovy_file_lines = f.readlines()

    def test_parsing_complicated_function(self):
        """ Groovy test parser functionality """
        groovy_file_def = parse(self.groovy_file)
        """ :type groovy_file_def: GroovyFileDef """
        print_(groovy_file_def)
        self.assertIsInstance(groovy_file_def, GroovyFileDef)
        self.assertEqual(groovy_file_def.filename, self.groovy_file)
        self.assertIsInstance(groovy_file_def.imports, (list, tuple))
        self.assertIsInstance(groovy_file_def.functions, (list, tuple))

        for import_def in groovy_file_def.imports:
            if import_def is not None:
                self.assertIsInstance(import_def, GroovyImport)
            print_("Import: {}".format(import_def))

        for func_def in groovy_file_def.functions:
            if func_def is not None:
                self.assertIsInstance(func_def, GroovyFunction)
            print_("Function: {}".format(func_def))

        groovy_funcs = groovy_file_def.functions

        self.assertEqual(len(groovy_funcs[6].body.split('\n')), 8)

        result_map = {x.name: x for x in groovy_funcs}
        self.assertIn('get_self', result_map)
        self.assertIn('return_value', result_map)
        self.assertIn('long_func', result_map)

    def test_parsing_example_function(self):
        test_func = 'def test_func(param1, param2) {\n    def n = null;\n    for (i in x) {\n n=i; \n}\n}\n'
        result = GroovyFunctionParser.parse(test_func)
        self.assertEqual(result.name, 'test_func')
        self.assertListEqual(result.args, ['param1', 'param2'])
        self.assertEqual(result.body, 'def n = null;\n    for (i in x) {\n n=i; \n}')
        self.assertEqual(result.defn, test_func)

    def test_parsing_example_import(self):
        test_import = 'import com.wellaware.test; //this is sample text'
        result = GroovyImportParser.parse(test_import)
        self.assertListEqual(result.comment_list, ['this', 'is', 'sample', 'text'])
        self.assertListEqual(result.import_list, ['import com.wellaware.test;'])
        self.assertListEqual(result.import_strings, ['com.wellaware.test'])

    def test_bad_parse(self):
        result = GroovyImportParser.parse(1)
        self.assertEqual(result, None)

        result = GroovyFunctionParser.parse(1)
        self.assertEqual(result, None)


from __future__ import unicode_literals
from mogwai._compat import print_, urllib
from mogwai.metrics.base import BaseMetricsReporter
import base64


class HostedGraphiteReporter(BaseMetricsReporter):
    """ Hosted Graphite Metrics Reporter class

    If you have a hosted graphite service, you can log the OGM metrics to your graphite service.
    """

    def __init__(self, api_key, url='https://hostedgraphite.com/api/v1/sink', *args, **kwargs):
        """ Create a Metrics Reporter

        To override this method, please make sure to call the superclass *after* your methods. Ie.
        def __init__(self, mystuff, stuff2=None, *args, **kwargs):
            # do my stuff...
            super(MyClass, self).__init__(*args, **kwargs)

        :param api_key: The Hosted Graphite API key
        :type api_key: basestring
        :param url: The Hosted Graphite API url
        :type url: basestring
        :param metric_prefix: The prefix on the collected metrics. The default implementation ignores this.
        :type metric_prefix: basestring | None
        :param registry: The pyformance registry
        :type registry: list[ MetricsRegistry | RegexRegistry ] | MetricsRegistry | RegexRegistry
        :param reportingInterval: The interval (number of seconds) on which to report the collected metrics.
        :type reportingInterval: float | long | int
        """
        self.url = url
        self.api_key = api_key
        super(HostedGraphiteReporter, self).__init__(*args, **kwargs)

    def send_metrics(self):
        """ Default Hosted Graphite implementation is to dump all metrics to STDOUT

        Change this if you want to customize for a particular metric collection system, ie. graphite, newrelic, etc.
        """
        metrics = self.get_metrics()
        if metrics:
            try:
                request = urllib.Request(self.url, metrics)
                request.add_header("Authorization", "Basic {}".format(base64.encodestring(self.api_key).strip()))
                result = urllib.urlopen(request)
            except Exception as e:
                print_(e)

    def get_metrics(self, timestamp=None):
        """ Default Hosted Graphite implementation to collect all the metrics from the registries and upload to a
        hosted graphite

        :param timestamp: use this timestamp instead of the generated timestamp when the method is run
        :type timestamp: long | int
        :returns: Timestamp and aggregated metrics from all registries.
        :rtype: tuple( long | int, dict )
        """
        timestamp, metrics = self._get_metrics(timestamp)
        metrics_data = []
        for key in metrics.keys():
            for valuekey in metrics[key].keys():
                metrics_data.append('{}_{}.{} {} {}\n'.format(self.metric_prefix,
                                                              key,
                                                              valuekey,
                                                              metrics[key][valuekey],
                                                              timestamp))
            return ''.join(metrics_data)



from __future__ import unicode_literals
from mogwai._compat import print_
from nose.plugins.attrib import attr

from mogwai.tests.base import BaseMogwaiTestCase, TestVertexModel, TestEdgeModel, TestEdgeModel2, TestEdgeModelDouble
from mogwai.exceptions import ValidationError, MogwaiQueryError


@attr('unit', 'edge_io')
class TestEdgeIO(BaseMogwaiTestCase):

    def setUp(self):
        super(TestEdgeIO, self).setUp()
        self.v1 = TestVertexModel.create(test_val=8, name='a')
        self.v2 = TestVertexModel.create(test_val=7, name='b')

    def tearDown(self):
        self.v1.delete()
        self.v2.delete()

    def test_model_save_and_load(self):
        """
        Tests that models can be saved and retrieved
        """
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)

        edges = self.v1.outE()
        self.assertEqual(len(edges), 1)
        self.assertEqual(edges[0].id, e1.id)

        e1.delete()

    def test_model_updating_works_properly(self):
        """
        Tests that subsequent saves after initial model creation work
        """
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)

        e1.test_val = 20
        e1.save()

        edges = self.v1.outE()
        self.assertEqual(len(edges), 1)
        self.assertEqual(edges[0].test_val, 20)

        e1.delete()

    def test_model_deleting_works_properly(self):
        """
        Tests that an instance's delete method deletes the instance
        """
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)

        e1.delete()
        edges = self.v1.outE()
        self.assertEqual(len(edges), 0)

        e1.delete()

        # don't actually create a db model and try to delete
        e2 = TestEdgeModel(self.v1, self.v2, test_val=3, name='nonexistant')
        e2.delete()

    def test_reload(self):
        """
        Tests that the reload method performs an inplace update of an instance's values
        """
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        e2 = TestEdgeModel.get(e1.id)
        print_('\n{} {} {}: {} {}'.format(e1.id, e2.id, e1 == e2, e1.test_val, e2.test_val))
        e2.test_val = 5
        e2.save()
        print_("{} {} {}: {} {}".format(e1.id, e2.id, e1 == e2, e1.test_val, e2.test_val))

        self.assertEqual(e1.test_val, 3)
        e1.reload()
        self.assertEqual(e1.test_val, 5)
        e1 = self.v1.outE()[0]
        self.assertEqual(e1.test_val, 5)

        e2.delete()

    def test_all_method_for_known_ids(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        e2 = TestEdgeModel.create(self.v1, self.v2, test_val=4)
        results = TestEdgeModel.all([e1.id, e2.id])
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIsInstance(result, TestEdgeModel)
            self.assertIn(result, [e1, e2])
        e1.delete()
        e2.delete()

    def test_all_method_for_known_bad_input(self):
        from mogwai.exceptions import MogwaiQueryError
        with self.assertRaises(TypeError):
            results = TestEdgeModel.all()
        with self.assertRaises(MogwaiQueryError):
            results = TestEdgeModel.all(None)
        with self.assertRaises(MogwaiQueryError):
            results = TestEdgeModel.all('test')

    def test_all_method_invalid_length(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        e2 = TestEdgeModel.create(self.v1, self.v2, test_val=4)
        from mogwai.exceptions import MogwaiQueryError
        with self.assertRaises(MogwaiQueryError):
            TestEdgeModel.all([e1.id, e2.id, 'invalid'])
        e1.delete()
        e2.delete()

    def test_in_between_method(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        results = TestEdgeModel.get_between(self.v1, self.v2)
        self.assertIsInstance(results, list)
        self.assertEqual(results[0], e1)
        e1.delete()

    def test_validation_error(self):
        from mogwai.exceptions import ValidationError
        with self.assertRaises(ValidationError):
            e1 = TestEdgeModel.create(self.v1, None, test_val=3)
        with self.assertRaises(ValidationError):
            e1 = TestEdgeModel.create(None, self.v2, test_val=3)
        with self.assertRaises(ValidationError):
            e1 = TestEdgeModel.create(None, None, test_val=3)

    def test_reload_no_changes(self):
        # NOTE titan 0.4.2 and earlier changes made to an edge deletes and then creates a new edge
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        e1.reload()
        e1.delete()

    def test_find_by_value_method(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=-99, name='e1')
        e2 = TestEdgeModel.create(self.v1, self.v2, test_val=-100, name='e2')
        e3 = TestEdgeModelDouble.create(self.v1, self.v2, test_val=-101.0, name='e3')

        self.assertEqual(len(TestEdgeModel.find_by_value('test_val', -99)), 1)
        self.assertEqual(len(TestEdgeModel.find_by_value('test_val', -100)), 1)
        self.assertEqual(len(TestEdgeModel.find_by_value('test_val', -102)), 0)
        self.assertEqual(len(TestEdgeModel.find_by_value('name', 'e2')), 1)
        self.assertEqual(len(TestEdgeModelDouble.find_by_value('test_val', -101.0)), 1)

        print_(e1)
        print_(e2)
        print_(e3)

        e1.delete()
        e2.delete()
        e3.delete()

    def test_get_by_id(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)
        results = TestEdgeModel.get(e1.id)
        self.assertIsInstance(results, TestEdgeModel)
        self.assertEqual(results, e1)

        from mogwai.exceptions import MogwaiQueryError
        with self.assertRaises(TestEdgeModel.DoesNotExist):
            results = TestEdgeModel.get(None)

        with self.assertRaises(TestEdgeModel.DoesNotExist):
            results = TestEdgeModel.get('nonexistant')

        e2 = TestEdgeModel2.create(self.v1, self.v2, test_val=2)
        with self.assertRaises(TestEdgeModel.WrongElementType):
            results = TestEdgeModel.get(e2.id)

        e1.delete()
        e2.delete()

    def test_inV_ouV_vertex_traversal(self):
        e1 = TestEdgeModel.create(self.v1, self.v2, test_val=3)

        v1 = e1.outV()
        v2 = e1.inV()
        self.assertEqual(v1, self.v1)
        self.assertEqual(v2, self.v2)

        e1.delete()



import logging
from mogwai._compat import array_types, integer_types, float_types, string_types, add_metaclass
from mogwai.connection import execute_query
from mogwai.exceptions import ElementDefinitionException, MogwaiQueryError, ValidationError
from mogwai.gremlin import GremlinMethod
from .element import Element, ElementMetaClass, edge_types

logger = logging.getLogger(__name__)


class EdgeMetaClass(ElementMetaClass):
    """Metaclass for edges."""

    def __new__(mcs, name, bases, body):
        #short circuit element_type inheritance
        body['label'] = body.pop('label', None)

        klass = super(EdgeMetaClass, mcs).__new__(mcs, name, bases, body)

        if not klass.__abstract__:
            label = klass.get_label()
            if label in edge_types and str(edge_types[label]) != str(klass):
                # Catch imports from other modules, don't reload module into vertex types, only load once
                logger.debug(ElementDefinitionException("%s is already registered as a edge: \n\tmcs: %s\n\tname: %s\n\tbases: %s\n\tbody: %s" % (label, mcs, name, bases, body)))
            else:
                edge_types[klass.get_label()] = klass
        return klass


@add_metaclass(EdgeMetaClass)
class Edge(Element):
    """Base class for all edges."""

    #__metaclass__ = EdgeMetaClass
    __abstract__ = True

    # if set to True, no more than one edge will
    # be created between two vertices
    __exclusive__ = False

    label = None

    gremlin_path = 'edge.groovy'

    _save_edge = GremlinMethod()
    _delete_edge = GremlinMethod()
    _get_edges_between = GremlinMethod(classmethod=True)
    _find_edge_by_value = GremlinMethod(classmethod=True)


    FACTORY_CLASS = None
    ## edge id
    ##edge_id = columns.UUID(save_strategy=columns.SAVE_ONCE)

    def __init__(self, outV, inV, **values):
        """
        Initialize this edge with the outgoing and incoming vertices as well as edge properties.

        :param outV: The vertex this edge is coming out of
        :type outV: Vertex
        :param inV: The vertex this edge is going into
        :type inV: Vertex
        :param values: The properties for this edge
        :type values: dict

        """
        self._outV = outV
        self._inV = inV
        super(Edge, self).__init__(**values)

    def __repr__(self):
        return "{}(label={}, id={}, values={})".format(self.__class__.__name__,
                                                       self.__class__.get_label(),
                                                       getattr(self, '_id', None),
                                                       getattr(self, '_values', {}))

    def __getstate__(self):
        state = {u'_id': self.id,
                 u'_type': u'edge',
                 u'_outV': str(self.outV().id),
                 u'_inV': str(self.inV().id),
                 u'_label': self.get_label(),
                 u'_properties': self.as_save_params()}
        return state

    def __setstate__(self, state):
        data = self.translate_db_fields(state)
        self.__init__(state['_outV'], state['_inV'], **data)
        return self

    @classmethod
    def find_by_value(cls, field, value, as_dict=False):
        """
        Returns edges that match the given field/value pair.

        :param field: The field to search
        :type field: str
        :param value: The value of the field
        :type value: str
        :param as_dict: Return results as a dictionary
        :type as_dict: boolean
        :rtype: [mogwai.models.Edge]
        """
        _field = cls.get_property_by_name(field)
        _label = cls.get_label()

        value_type = False
        if isinstance(value, integer_types + float_types):
            value_type = True

        results = cls._find_edge_by_value(
            value_type=value_type,
            label=_label,
            field=_field,
            value=value
        )

        if as_dict:  # pragma: no cover
            return {v._id: v for v in results}

        return results

    @classmethod
    def all(cls, ids, as_dict=False):
        """
        Load all edges with the given edge_ids from the graph. By default this will return a list of edges but if
        as_dict is True then it will return a dictionary containing edge_ids as keys and edges found as values.

        :param ids: A list of titan IDs
        :type ids: list
        :param as_dict: Toggle whether to return a dictionary or list
        :type as_dict: boolean
        :rtype: dict | list
        """
        if not isinstance(ids, array_types):
            raise MogwaiQueryError("ids must be of type list or tuple")

        strids = [str(i) for i in ids]
        qs = ['ids.collect{g.e(it)}']

        results = execute_query('\n'.join(qs), {'ids': strids})
        results = list(filter(None, results))

        if len(results) != len(ids):
            raise MogwaiQueryError("the number of results don't match the number of edge ids requested")

        objects = []
        for r in results:
            try:
                objects += [Element.deserialize(r)]
            except KeyError:  # pragma: no cover
                raise MogwaiQueryError('Edge type "%s" is unknown' % '')

        if as_dict:  # pragma: no cover
            return {e._id: e for e in objects}

        return objects

    @classmethod
    def get_label(cls):
        """
        Returns the label for this edge.

        :rtype: str

        """
        return cls._type_name(cls.label)

    @classmethod
    def get_between(cls, outV, inV, page_num=None, per_page=None):
        """
        Return all the edges with a given label between two vertices.

        :param outV: The vertex the edge comes out of.
        :type outV: Vertex
        :param inV: The vertex the edge goes into.
        :type inV: Vertex
        :param page_num: The page number of the results
        :type page_num: int
        :param per_page: The number of results per page
        :type per_page: int
        :rtype: list

        """
        return cls._get_edges_between(out_v=outV,
                                      in_v=inV,
                                      label=cls.get_label(),
                                      page_num=page_num,
                                      per_page=per_page)

    def validate(self):
        """
        Perform validation of this edge raising a ValidationError if any problems are encountered.
        """
        if self._id is None:
            if self._inV is None:
                raise ValidationError('in vertex must be set before saving new edges')
            if self._outV is None:
                raise ValidationError('out vertex must be set before saving new edges')
        super(Edge, self).validate()

    def save(self, *args, **kwargs):
        """
        Save this edge to the graph database.
        """
        super(Edge, self).save(*args, **kwargs)
        return self._save_edge(self._outV,
                               self._inV,
                               self.get_label(),
                               self.as_save_params(),
                               exclusive=self.__exclusive__)

    def _reload_values(self):
        """ Re-read the values for this edge from the graph database. """
        reloaded_values = {}
        results = execute_query('g.e(id)', {'id': self._id})
        if results:  # note this won't work if you update a node for titan pre-0.5.x, new id's are created
            #del results['_id']
            del results['_type']
            reloaded_values['_id'] = results['_id']
            for name, value in results.get('_properties', {}).items():
                reloaded_values[name] = value
            if results['_id']:
                setattr(self, '_id', results['_id'])
            return reloaded_values
        else:
            return {}

    @classmethod
    def get(cls, id):
        """
        Look up edge by titan assigned ID. Raises a DoesNotExist exception if a edge with the given edge id was not
        found. Raises a MultipleObjectsReturned exception if the edge_id corresponds to more than one edge in the graph.

        :param id: The titan assigned ID
        :type id: str | basestring
        :rtype: mogwai.models.Edge
        """
        try:
            results = cls.all([id])
            if len(results) > 1:  # pragma: no cover
                # This requires titan to be broken.
                raise cls.MultipleObjectsReturned

            result = results[0]
            if not isinstance(result, cls):
                raise cls.WrongElementType(
                    '%s is not an instance or subclass of %s' % (result.__class__.__name__, cls.__name__)
                )
            return result
        except MogwaiQueryError:
            raise cls.DoesNotExist

    @classmethod
    def create(cls, outV, inV, label=None, *args, **kwargs):
        """
        Create a new edge of the current type coming out of vertex outV and going into vertex inV with the given
        properties.

        :param outV: The vertex the edge is coming out of
        :type outV: Vertex
        :param inV: The vertex the edge is going into
        :type inV: Vertex

        """
        return super(Edge, cls).create(outV, inV, *args, **kwargs)

    def delete(self):
        """
        Delete the current edge from the graph.
        """
        if self.__abstract__:  # pragma: no cover
            raise MogwaiQueryError('cant delete abstract elements')
        if self._id is None:
            return self
        self._delete_edge()

    def _simple_traversal(self, operation):
        """
        Perform a simple traversal starting from the current edge returning a list of results.

        :param operation: The operation to be performed
        :type operation: str
        :rtype: list

        """
        results = execute_query('g.e(id).%s()' % operation, {'id': self.id})
        return [Element.deserialize(r) for r in results]

    def inV(self):
        """
        Return the vertex that this edge goes into.

        :rtype: Vertex

        """
        from mogwai.models.vertex import Vertex

        if self._inV is None:
            self._inV = self._simple_traversal('inV')[0]
        if isinstance(self._inV, string_types + integer_types):
            self._inV = Vertex.get(self._inV)
        return self._inV

    def outV(self):
        """
        Return the vertex that this edge is coming out of.

        :rtype: Vertex

        """
        from mogwai.models.vertex import Vertex

        if self._outV is None:
            self._outV = self._simple_traversal('outV')[0]
        if isinstance(self._inV, string_types + integer_types):
            self._outV = Vertex.get(self._outV)
        return self._outV



from .vertex import Vertex
from .edge import Edge
from .paginated_vertex import PaginatedVertex
from .query import Query

from mogwai.constants import EQUAL, GREATER_THAN_EQUAL, GREATER_THAN, \
    LESS_THAN_EQUAL, LESS_THAN, NOT_EQUAL, OUT, IN, BOTH



from __future__ import unicode_literals
from mogwai._compat import string_types, PY2
from nose.plugins.attrib import attr
from mogwai.tests import BaseMogwaiTestCase
from mogwai.models import Vertex, Edge
from mogwai import properties
from mogwai.tools import ImportStringError, import_string, cached_property, LazyImportClass, _Missing, Factory
import factory


@attr('unit', 'tools')
class TestImportString(BaseMogwaiTestCase):

    def test_raises_import_string_error(self):
        with self.assertRaises(ImportStringError):
            import_string('does.not.exist')
        with self.assertRaises(ImportStringError):
            import_string('does.not:exist')

    def test_imports_correctly(self):
        cp = import_string('mogwai.tools:cached_property')
        self.assertIs(cp, cached_property)

    def test_import_string_error_class(self):
        ise = ImportStringError('mogwai.tools.cached_property', 'Should work')


class LazyImportObject(object):
    pass


@attr('unit', 'tools')
class TestLazyImportString(BaseMogwaiTestCase):

    def test_lazy_import(self):
        li = LazyImportClass('mogwai.tools._Missing')
        self.assertIsInstance(li, LazyImportClass)
        self.assertIs(li.klass, _Missing)
        self.assertIsInstance(li(), _Missing)


class TestFactoryVertex(Vertex):
    element_type = 'test_factory_vertex'

    name = properties.String(required=True, max_length=128)


class TestFactoryVertexFactory(Factory):
    FACTORY_FOR = TestFactoryVertex

    name = factory.Sequence(lambda n: 'Test Factory Vertex {}'.format(n))


class TestFactoryEdge(Edge):
    label = 'test_factory_edge'

    name = properties.String(required=True, max_length=128)


class TestFactoryEdgeFactory(Factory):
    FACTORY_FOR = TestFactoryEdge

    name = factory.Sequence(lambda n: 'Test Factory Edge {}'.format(n))


@attr('unit', 'tools', 'factory')
class TestModelFactory(BaseMogwaiTestCase):

    def test_vertex_factory(self):
        v1 = TestFactoryVertexFactory()
        self.assertIsInstance(v1, TestFactoryVertex)
        self.assertTrue(v1.name.startswith('Test Factory Vertex'))
        v1.delete()
        v2 = TestFactoryVertexFactory.create()
        self.assertIsInstance(v2, TestFactoryVertex)
        self.assertTrue(v2.name.startswith('Test Factory Vertex'))
        v2.delete()

    def test_edge_factory(self):
        v1 = TestFactoryVertexFactory()
        v2 = TestFactoryVertexFactory()
        e1 = TestFactoryEdgeFactory(outV=v1, inV=v2)
        e2 = TestFactoryEdgeFactory.create(outV=v1, inV=v2)
        self.assertIsInstance(e1, TestFactoryEdge)
        self.assertTrue(e1.name.startswith('Test Factory Edge'))
        self.assertIsInstance(e2, TestFactoryEdge)
        self.assertTrue(e2.name.startswith('Test Factory Edge'))
        e1.delete()
        e2.delete()
        v1.delete()
        v2.delete()


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from .base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import IPV4
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_ipv4')
class IPV4PropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = IPV4
    good_cases = ('1.2.3.4', '0.0.0.0', '255.255.255.255')
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256')

    def test_ipv4_default_cases(self):
        p = IPV4(default='1.2.3.4')


class IPV4TestVertex(Vertex):
    element_type = 'test_ipv4_vertex'

    test_val = IPV4()


@attr('unit', 'property', 'property_ipv4')
class IPV4VertexTestCase(GraphPropertyBaseClassTestCase):

    def test_ipv4_io(self):
        print_("creating vertex")
        dt = IPV4TestVertex.create(test_val='1.2.3.4')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = IPV4TestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = IPV4TestVertex.create(test_val='4.3.2.1')
        print_("\ncreated vertex: %s" % dt)
        dt2 = IPV4TestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, '4.3.2.1')
        print_("deleting vertex")
        dt2.delete()


import inspect
import os.path
from datetime import datetime
from decimal import Decimal as _Decimal
from uuid import UUID as _UUID
import logging
from mogwai._compat import array_types, string_types, integer_types, float_types, iteritems
from mogwai.connection import execute_query
from mogwai.exceptions import MogwaiQueryError, MogwaiGremlinException
from .groovy import parse, GroovyImport
from .table import Table, Row

logger = logging.getLogger(__name__)


class BaseGremlinMethod(object):
    """ Maps a function in a groovy file to a method on a python class """

    def __init__(self,
                 path=None,
                 method_name=None,
                 classmethod=False,
                 property=False,
                 defaults={},
                 transaction=True,
                 imports=None):
        """
        Initialize the gremlin method and define how it is attached to class.

        :param path: Path to the gremlin source (relative to file class is defined in). Absolute paths work as well.
                     Defaults to gremlin.groovy.
        :type path: str
        :param method_name: The name of the function definition in the groovy file
        :type method_name: str
        :param classmethod: Method should behave as a classmethod if True
        :type classmethod: bool
        :param property: Method should behave as a property
        :type property: bool
        :param defaults: The default parameters to the function
        :type defaults: dict
        :param transaction: Close previous transaction before executing (True by default)
        :type transaction: bool
        :param imports: Additional imports to include when calling the GremlinMethod
        :type imports: list | tuple | str

        """
        self.is_configured = False
        self.is_setup = False
        self.path = path
        self.method_name = method_name
        self.classmethod = classmethod
        self.property = property
        self.defaults =defaults
        self.transaction = transaction

        # function
        self.attr_name = None
        self.arg_list = []
        self.function_body = None
        self.function_def = None

        # imports
        self.imports = None
        imports = imports or []
        if isinstance(imports, (str, string_types)):
            imports = [imports, ]
        self.extra_imports = imports

        #configuring attributes
        self.parent_class = None

    def configure_method(self, klass, attr_name, gremlin_path):
        """
        Sets up the methods internals

        :param klass: The class object this function is being added to
        :type klass: object
        :param attr_name: The attribute name this function will be added as
        :type attr_name: str
        :param gremlin_path: The path to the gremlin file containing method
        :type gremlin_path: str

        """
        if not self.is_configured:
            self.parent_class = klass
            self.attr_name = attr_name
            self.method_name = self.method_name or self.attr_name
            self.path = self.path or gremlin_path

            self.is_configured = True

    def _setup(self):
        """
        Does the actual method configuration, this is here because the method configuration must happen after the class
        is defined.
        """
        if not self.is_setup:

            #construct the default name
            name_func = getattr(self.parent_class, 'get_element_type', None) or \
                        getattr(self.parent_class, 'get_label', None)
            default_path = (name_func() if name_func else 'gremlin') + '.groovy'

            self.path = self.path or default_path
            if self.path.startswith('/'):
                path = self.path  # pragma: no cover
            else:
                path = inspect.getfile(self.parent_class)
                path = os.path.split(path)[0]
                path += '/' + self.path

            #TODO: make this less naive
            gremlin_obj = None
            file_def = parse(path)

            # Functions
            for grem_func in file_def.functions:
                if grem_func.name == self.method_name:
                    gremlin_obj = grem_func
                    break

            if gremlin_obj is None:
                raise MogwaiGremlinException("The method '%s' wasn't found in %s" % (self.method_name, path))

            for arg in gremlin_obj.args:
                if arg in self.arg_list:
                    raise MogwaiGremlinException("'%s' defined more than once in gremlin method arguments" % arg)
                self.arg_list.append(arg)

            self.function_body = gremlin_obj.body
            self.function_def = gremlin_obj.defn

            # imports
            self.imports = file_def.imports
            extra_imports = []
            for extra_import in self.extra_imports:
                extra_imports.append(GroovyImport([], [extra_import], ['import {};'.format(extra_import)]))
            self.extra_imports = extra_imports

            self.is_setup = True

    def __call__(self, instance, *args, **kwargs):
        """
        Intercept attempts to call the GremlinMethod attribute and perform a gremlin query returning the results.

        :param instance: The class instance the method was called on
        :type instance: object

        """
        self._setup()

        args = list(args)
        if not self.classmethod:
            args = [instance._id] + args

        params = self.defaults.copy()
        if len(args + list(kwargs.values())) > len(self.arg_list):  # pragma: no cover
            raise TypeError('%s() takes %s args, %s given' % (self.attr_name, len(self.arg_list), len(args)))

        #check for and calculate callable defaults
        for k, v in params.items():
            if callable(v):
                params[k] = v()

        arglist = self.arg_list[:]
        for arg in args:
            params[arglist.pop(0)] = arg

        for k, v in kwargs.items():
            if k not in arglist:
                an = self.attr_name
                if k in params:  # pragma: no cover
                    raise TypeError("%s() got multiple values for keyword argument '%s'" % (an, k))
                else:  # pragma: no cover
                    raise TypeError("%s() got an unexpected keyword argument '%s'" % (an, k))
            arglist.pop(arglist.index(k))
            params[k] = v

        params = self.transform_params_to_database(params)

        import_list = []
        for imp in self.imports + self.extra_imports:
            if imp is not None:
                for import_string in imp.import_list:
                    import_list.append(import_string)
        import_string = '\n'.join(import_list)

        script = '\n'.join([import_string, self.function_body])

        try:
            if hasattr(instance, 'get_element_type'):
                context = "vertices.{}".format(instance.get_element_type())
            elif hasattr(instance, 'get_label'):
                context = "edges.{}".format(instance.get_label())
            else:
                context = "other"

            context = "{}.{}".format(context, self.method_name)

            tmp = execute_query(script, params, transaction=self.transaction, context=context)
        except MogwaiQueryError as pqe:  # pragma: no cover
            import pprint
            msg = "Error while executing Gremlin method\n\n"
            msg += "[Method]\n{}\n\n".format(self.method_name)
            msg += "[Params]\n{}\n\n".format(pprint.pformat(params))
            msg += "[Function Body]\n{}\n".format(self.function_body)
            msg += "[Imports]\n{}\n".format(import_string)
            msg += "\n[Error]\n{}\n".format(pqe)
            if hasattr(pqe, 'raw_response'):
                msg += "\n[Raw Response]\n{}\n".format(pqe.raw_response)
            raise MogwaiGremlinException(msg)
        return tmp

    def transform_params_to_database(self, params):
        """
        Takes a dictionary of parameters and recursively translates them into parameters appropriate for sending over
        Rexpro.

        :param params: The parameters to be sent to the function
        :type params: dict
        :rtype: dict

        """
        from mogwai.models.element import BaseElement
        from mogwai.models import Edge, Vertex
        from mogwai.properties import DateTime, Decimal, UUID

        if isinstance(params, dict):
            return {k: self.transform_params_to_database(v) for k, v in iteritems(params)}
        if isinstance(params, array_types):
            return [self.transform_params_to_database(x) for x in params]
        if isinstance(params, BaseElement):
            return params._id
        if inspect.isclass(params) and issubclass(params, Edge):
            return params.label
        if inspect.isclass(params) and issubclass(params, Vertex):
            return params.element_type
        if isinstance(params, datetime):
            return DateTime().to_database(params)
        if isinstance(params, _UUID):
            return UUID().to_database(params)
        if isinstance(params, _Decimal):
            return Decimal().to_database(params)
        return params


class GremlinMethod(BaseGremlinMethod):
    """Gremlin method that returns a graph element"""

    @staticmethod
    def _deserialize(obj):
        """
        Recursively deserializes elements returned from rexster

        :param obj: The raw result returned from rexster
        :type obj: object

        """
        from mogwai.models.element import Element

        if isinstance(obj, dict) and '_id' in obj and '_type' in obj:
            return Element.deserialize(obj)
        elif isinstance(obj, dict):
            return {k: GremlinMethod._deserialize(v) for k, v in obj.items()}
        elif isinstance(obj, array_types):
            return [GremlinMethod._deserialize(v) for v in obj]
        else:
            return obj

    def __call__(self, instance, *args, **kwargs):
        results = super(GremlinMethod, self).__call__(instance, *args, **kwargs)
        return GremlinMethod._deserialize(results)


class GremlinValue(GremlinMethod):
    """Gremlin Method that returns one value"""

    def __call__(self, instance, *args, **kwargs):
        results = super(GremlinValue, self).__call__(instance, *args, **kwargs)

        if results is None:  # pragma: no cover
            return

        # we have to make a special case for dictionaries, python len returns number of keys,
        # even though it is one object. Don't do the same for lists or tuples, since they are
        # Gremlin Methods not GremlinValues.
        if isinstance(results, dict):
            return results
        if isinstance(results, integer_types + float_types + string_types):
            return results
        from mogwai.models.element import Element
        if isinstance(results, Element):
            return results

        if len(results) != 1:
            raise MogwaiGremlinException('GremlinValue requires a single value is returned (%s returned)' % len(results))
        return results[0]


class GremlinTable(GremlinMethod):  # pragma: no cover
    """Gremlin method that returns a table as its result"""

    def __call__(self, instance, *args, **kwargs):
        results = super(GremlinTable, self).__call__(instance, *args, **kwargs)
        if results is None or (isinstance(results, array_types) and len(results) != 1):
            return
        return Table(results[0])



from __future__ import unicode_literals
from mogwai.connection import setup, _parse_host, execute_query
from mogwai import models
from mogwai import properties
from mogwai import gremlin
from mogwai import exceptions

import readline
import code
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--host", help="Enter the hostname of the graph database")
parser.add_argument("--username", help="Enter the username associated with the graph database")
parser.add_argument("--password", help="Enter the password associated with the graph database")
parser.add_argument("--database", help="Enter the graph database name")
args = parser.parse_args()

if not args.host:
    args.host = 'localhost'
if not args.username:
    args.username = ''
if not args.password:
    args.password = ''
if not args.database:
    args.database = 'graph'

hd = _parse_host(args.host, args.username, args.password, args.database)
host = hd.get('host', 'localhost')
port = hd.get('port', 8184)
username = hd.get('username', '')
password = hd.get('password', '')
graph_name = hd.get('graph_name', 'graph')

setup([host, ], graph_name=graph_name, username=username, password=password)

del args, parser, username, password, hd


class __help_object(object):

    def __repr__(self):
        return """The following have already been imported:
  - mogwai.execute_query as execute_query
  - mogwai.models as models
  - mogwai.properties as properties
  - mogwai.gremlin as gremlin
  - mogwai.exceptions as exceptions"""

    def __str__(self):
        return self.__repr__()

    def __unicode__(self):
        return self.__repr__()

HELP = __help_object()

vars = globals().copy()
vars.update(locals())

shell = code.InteractiveConsole(vars)
shell.interact(banner="Mogwai GraphDB Shell. Connected to %s:%s on graph `%s`. Type 'HELP' for available commands"
                      % (host, port, graph_name))



from __future__ import unicode_literals
from mogwai._compat import print_
from six import string_types
from nose.plugins.attrib import attr
from nose.tools import nottest


from .base import BaseMogwaiTestCase
from mogwai.connection import generate_spec, execute_query
from mogwai.models import Vertex, Edge
from mogwai.properties import String
from mogwai.spec import get_existing_indices


class TestIndexSpecVertex(Vertex):
    element_type = 'test_index_spec_vertex_model'

    name = String(default='test_vertex', index=True, index_ext='es')


class TestIndexSpecEdge(Edge):
    label = 'test_index_spec_edge_model'

    name = String(default='test_edge', index=True, index_ext='es')


@attr('unit', 'connection')
class TestSpecSystem(BaseMogwaiTestCase):
    """ Test specification system """

    def test_loaded(self):
        spec = generate_spec()
        self.assertIsInstance(spec, (list, tuple))
        self.assertGreater(len(spec), 0)
        for s in spec:
            print_(s)
            self.assertIsInstance(s, dict)
            self.assertDictContainsKeyWithValueType(s, 'model', string_types)
            self.assertDictContainsKeyWithValueType(s, 'element_type', string_types)
            self.assertDictContainsKeyWithValueType(s, 'makeType', string_types)
            self.assertDictContainsKeyWithValueType(s, 'properties', dict)
            for pk, pv in s['properties'].items():
                self.assertDictContainsKeyWithValueType(pv, 'data_type', string_types)
                self.assertDictContainsKeyWithValueType(pv, 'index_ext', string_types)
                self.assertDictContainsKeyWithValueType(pv, 'uniqueness', string_types)
                self.assertDictContainsKeyWithValueType(pv, 'compiled', dict)
                self.assertDictContainsKeyWithValueType(pv['compiled'], 'script', string_types)
                self.assertDictContainsKeyWithValueType(pv['compiled'], 'params', dict)
                self.assertDictContainsKeyWithValueType(pv['compiled'], 'transaction', bool)

    @nottest
    def test_gather_existing_indices(self):
        """ Make sure existing indices can be gathered """
        v_idx, e_idx = get_existing_indices()
        self.assertEqual(len(v_idx), 0)
        self.assertEqual(len(e_idx), 0)

        # create vertex and edge index
        execute_query('g.makeKey(name).dataType(Object.class).indexed(Vertex.class).make(); g.commit()',
                      params={'name': 'testvertexindex'})
        execute_query('g.makeLabel(name).dataType(Object.class).indexed(Edge.class).make(); g.commit()',
                      params={'name': 'testedgeindex'})
        v_idx, e_idx = get_existing_indices()
        self.assertEqual(len(v_idx), 1)
        self.assertEqual(len(e_idx), 1)


import sys
from setuptools import setup, find_packages

#next time:
#python setup.py register
#python setup.py sdist upload

version = open('mogwai/VERSION', 'r').readline().strip()
develop_requires = ['Sphinx==1.2.2',
    'Twisted>=13.2.0,<14.0.0',
    'blinker==1.2',
    'coverage==3.7.1',
    'detox==0.9.3',
    'docutils==0.11',
    'eventlet>=0.14.0',
    'factory-boy==2.4.1',
    'gevent>=1.0.1',
    'msgpack-python==0.4.2',
    'newrelic==2.18.1.15',
    'nose==1.3.0',
    'pyformance==0.2.4',
    'pyparsing==2.0.2',
    'pytz==2014.4',
    'rexpro>=0.3.0,<1.0.0',
    'six>=1.6.1',
    'sphinx-rtd-theme==0.1.6',
    'tox==1.7.1',
    'watchdog==0.7.1',
    'wercker==0.8.3',
    'wheel>=0.24.0',
    'pyxdg>=0.25',
    'keyring>=4.0']

long_desc = """
mogwai is an Object-Graph Mapper (OGM) for Python

`Documentation <https://mogwai.readthedocs.org/en/latest/>`_

`Report a Bug <https://bitbucket.org/wellaware/mogwai/issues>`_
"""

setup(
    name='mogwai',
    version=version,
    description='Titan Object-Graph Mapper (OGM)',
    dependency_links=['https://bitbucket.org/wellaware/mogwai/archive/{0}.tar.gz#egg=mogwai-{0}'.format(version)],
    long_description=long_desc,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Environment :: Other Environment",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation",
        "Topic :: Database",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='cassandra,titan,ogm,mogwai,thunderdome',
    install_requires=['pyparsing>=2.0.2,<3.0.0',
                      'rexpro>=0.3.3,<1.0.0',
                      'six==1.6.1',
                      'argparse>=1.2.1',
                      'factory-boy>=2.4.1',
                      'pyformance==0.2.4',
                      'Twisted>=13.2.0,<14.0.0',
                      'pytz>=2014.4'],
    extras_require={
        'develop': develop_requires,
        'newrelic': ['newrelic>=2.18.1.15'],
        'docs': ['Sphinx>=1.2.2', 'sphinx-rtd-theme>=0.1.6', 'watchdog>=0.7.1', 'newrelic>=2.18.1.15'],
        'gevent': ['rexpro[gevent]>=0.3.3'],
        'eventlet': ['rexpro[eventlet]>=0.3.3'],
    },
    test_suite='nose.collector',
    tests_require=develop_requires,
    author='Cody Lee',
    author_email='codylee@wellaware.us',
    maintainer='Cody Lee',
    maintainer_email='codylee@wellaware.us',
    url='https://bitbucket.org/wellaware/mogwai',
    license='Apache Software License 2.0',
    packages=find_packages(),
    include_package_data=True,
)



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from .base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Dictionary
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_dict')
class DictionaryPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Dictionary
    good_cases = ({}, {'test': 1}, None)
    bad_cases = (0, 1.1, 'val', ['val'])


class DictionaryTestVertex(Vertex):
    element_type = 'test_dictionary_vertex'

    test_val = Dictionary()


@attr('unit', 'property', 'property_dict')
class DictionaryVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_dictionary_io(self):
        print_("creating vertex")
        dt = DictionaryTestVertex.create(test_val={'test': 1})
        print_("getting vertex from vertex: %s" % dt)
        dt2 = DictionaryTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = DictionaryTestVertex.create(test_val={'test2': 2})
        print_("\ncreated vertex: %s" % dt)
        dt2 = DictionaryTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, {'test2': 2})
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
import six

PY2 = six.PY2
PY3 = six.PY3

# conversions
unichr = six.unichr
if PY3:
    long_ = int
else:
    long_ = long
text_type = six.text_type
string_types = six.string_types
integer_types = six.integer_types
bool_types = (bool, )
binary_types = (six.binary_type, )
float_types = (float, )
array_types = (tuple, list)
int_to_byte = six.int2byte

# iterator functions
iterkeys = six.iterkeys
itervalues = six.itervalues
iteritems = six.iteritems
iterlists = six.iterlists
iterbytes = six.iterbytes
reraise = six.reraise

with_metaclass = six.with_metaclass
add_metaclass = six.add_metaclass
print_ = six.print_
urllib = six.moves.urllib

get_method_self = six.get_method_self


from __future__ import unicode_literals
from collections import OrderedDict
import re

from mogwai._compat import string_types, print_, add_metaclass
from mogwai.tools import import_string
from mogwai import properties
from mogwai.exceptions import MogwaiException, SaveStrategyException, \
    ModelException, ElementDefinitionException
from mogwai.gremlin import BaseGremlinMethod
from mogwai.connection import _add_model_to_space

# import for backward compatibility
from mogwai.constants import BOTH, EQUAL, GREATER_THAN, GREATER_THAN_EQUAL, \
    IN, LESS_THAN, LESS_THAN_EQUAL, NOT_EQUAL, OUT


import logging
import warnings
logger = logging.getLogger(__name__)

#dict of node and edge types for rehydrating results
vertex_types = {}
edge_types = {}


class BaseElement(object):
    """
    The base model class, don't inherit from this, inherit from Model, defined
    below
    """
    #__enum_id_only__ = True
    FACTORY_CLASS = None

    class DoesNotExist(MogwaiException):
        """
        Object not found in database
        """
        pass

    class MultipleObjectsReturned(MogwaiException):
        """
        Multiple objects returned on unique key lookup
        """
        pass

    class WrongElementType(MogwaiException):
        """
        Unique lookup with key corresponding to vertex of different type
        """
        pass

    def __init__(self, **values):
        """
        Initialize the element with the given properties.

        :param values: The properties for this element
        :type values: dict

        """
        self._id = values.get('_id')
        self._values = {}
        self._manual_values = {}
        #print_("Received values: %s" % values)
        #print_("Known Relationships: %s" % self._relationships)
        for name, prop in self._properties.items():
            #print_("trying name: %s in values" % name)
            value = values.get(name, None)
            if value is not None:
                #print_("Got value")
                value = prop.to_python(value)
            #else:
                #print_("no value found")
            value_mngr = prop.value_manager(prop, value, prop.save_strategy)
            self._values[name] = value_mngr
            setattr(self, name, value)
        for name, relationship in self._relationships.items():
            relationship._setup_instantiated_vertex(self)

        # unknown properties that are loaded manually
        from mogwai.properties.base import BaseValueManager
        for kwarg in set(values.keys()).difference(set(self._properties.keys())):  # set(self._properties.keys()) - set(values.keys()):
            if kwarg not in ('_id', '_inV', '_outV', 'element_type'):
                self._manual_values[kwarg] = BaseValueManager(None, values.get(kwarg))

    @property
    def id(self):
        return self._id

    def __eq__(self, other):
        """
        Check for equality between two elements.

        :param other: Element to be compared to
        :type other: BaseElement
        :rtype: boolean

        """
        if not isinstance(other, BaseElement):  # pragma: no cover
            return False
        return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        """
        Check for inequality between two elements.

        :param other: Element to be compared to
        :type other: BaseElement
        :rtype: boolean

        """
        return not self.__eq__(other)  # pragma: no cover

    @classmethod
    def _type_name(cls, manual_name):
        """
        Returns the element name if it has been defined, otherwise it creates it from the module and class name.

        :param manual_name: Name to override the default type name
        :type manual_name: str
        :rtype: str

        """
        pf_name = ''
        if manual_name:
            pf_name = manual_name.lower()
        else:
            camelcase = re.compile(r'([a-z])([A-Z])')
            ccase = lambda s: camelcase.sub(lambda v: '{}_{}'.format(v.group(1), v.group(2).lower()), s)

            pf_name += ccase(cls.__name__)
            pf_name = pf_name.lower()
        return pf_name.lstrip('_')

    def validate_field(self, field_name, val):
        """
        Perform the validations associated with the field with the given name on the value passed.

        :param field_name: The name of property whose validations will be run
        :type field_name: str
        :param val: The value to be validated
        :type val: mixed

        """
        return self._properties[field_name].validate(val)

    def validate(self):
        """Cleans and validates the field values"""
        for name in self._properties.keys():
            #print_("Validating {}...".format(name))
            func_name = 'validate_{}'.format(name)
            val = getattr(self, name)
            #print_("Got {}: func_name: {}, attr: {} ({})".format(name, func_name, val, type(val)))
            if hasattr(self, func_name):
                #print_("Calling custom function: {}".format(func_name))
                val = getattr(self, func_name)(val)
            else:
                #print_("Calling standard function: {}".format(self._properties[name]))
                val = self._properties[name].validate(val)  # self.validate_field(name, val)
            #print_("Validated {}: val: {} ({}), func_name: {}".format(name, val, type(val), func_name))
            setattr(self, name, val)

    def as_dict(self):
        """
        Returns a map of column names to cleaned values

        :rtype: dict

        """
        values = {}
        for name, prop in self._properties.items():
            values[name] = prop.to_database(getattr(self, name, None))
        values.update(self._manual_values)
        values['id'] = self.id
        return values

    def as_save_params(self):
        """
        Returns a map of property names to cleaned values containing only the properties which should be persisted on
        save.

        :rtype: dict

        """
        values = {}
        was_saved = self._id is not None
        for name, prop in self._properties.items():
            # Determine the save strategy for this column
            prop_strategy = prop.get_save_strategy()

            # Enforce the save strategy
            vm = self._values[name]
            should_save = prop_strategy.condition(previous_value=vm.previous_value,
                                                  value=vm.value,
                                                  has_changed=vm.changed,
                                                  first_save=was_saved,
                                                  graph_property=prop)

            if should_save:
                #print_("Saving %s to database for name %s" % (prop.db_field_name or name, name))
                values[prop.db_field_name or name] = prop.to_database(vm.value)

        # manual values
        for name, prop in self._manual_values.items():
            if prop is None:
                # Remove this property entirely
                values[name] = None
            else:
                # Determine the save strategy
                prop_strategy = prop.strategy
                if prop_strategy.condition(previous_value=prop.previous_value,
                                           value=prop.value,
                                           has_changed=prop.changed,
                                           first_save=was_saved,
                                           graph_property=None):
                    values[name] = prop.value

        return values

    @classmethod
    def translate_db_fields(cls, data):
        """
        Translates field names from the database into field names used in our model

        this is for cases where we're saving a field under a different name than it's model property

        :param data: dict
        :rtype: dict
        """
        dst_data = data.copy().get('_properties', {})
        if data.get('_id', None):
            dst_data.update({'_id': data.copy().get('_id')})
        #print_("Raw incoming data: %s" % data)
        for name, prop in cls._properties.items():
            #print_("trying db_field_name: %s and name: %s" % (prop.db_field_name, name))
            if prop.db_field_name in dst_data:
                dst_data[name] = dst_data.pop(prop.db_field_name)
            elif name in dst_data:
                dst_data[name] = dst_data.pop(name)

        return dst_data

    @classmethod
    def create(cls, *args, **kwargs):
        """Create a new element with the given information."""
        return cls(*args, **kwargs).save()

    def pre_save(self):
        """Pre-save hook which is run before saving an element"""
        self.validate()

    def save(self):
        """
        Base class save method. Performs basic validation and error handling.
        """
        if self.__abstract__:
            raise MogwaiException('cant save abstract elements')
        self.pre_save()
        return self

    def pre_update(self, **values):
        """ Override this to perform pre-update validation """
        pass

    def update(self, **values):
        """
        performs an update of this element with the given values and returns the
        saved object
        """
        if self.__abstract__:
            raise MogwaiException('cant update abstract elements')
        self.pre_update(**values)
        for key in values.keys():
            if key not in self._properties:
                raise TypeError("unrecognized attribute name: '{}'".format(key))

        for k, v in values.items():
            setattr(self, k, v)

        return self.save()

    def _reload_values(self):
        """
        Base method for reloading an element from the database.

        """
        raise NotImplementedError

    def reload(self):
        """
        Reload the given element from the database.

        """
        values = self._reload_values()
        for name, prop in self._properties.items():
            value = values.get(prop.db_field_name, None)
            if value is not None:
                value = prop.to_python(value)
            setattr(self, name, value)
        return self

    @classmethod
    def get_property_by_name(cls, key):
        """
        Get's the db_field_name of a property by key

        :param key: attribute of the model
        :type key: basestring | str
        :rtype: basestring | str | None
        """
        if isinstance(key, string_types):
            prop = cls._properties.get(key, None)
            if prop:
                return prop.db_field_name
        return key  # pragma: no cover

    @classmethod
    def _get_factory(cls):
        if getattr(cls, 'FACTORY_CLASS', None):
            factory_cls = getattr(cls, 'FACTORY_CLASS', None)
            if isinstance(factory_cls, string_types):
                factory_cls = import_string(factory_cls)
            create_cls = factory_cls.create
        else:
            create_cls = cls.create
        
        return create_cls

    def __getitem__(self, item):
        logger.debug("Using __getitem__({})...".format(item))
        #print_("Using __getitem__({})...: options: _properties({}), _manual_values({})".format(item, self._properties.items(), self._manual_values.items()))
        value = self._properties.get(item, None)
        if value is not None:
            # call the normal getattr method
            return getattr(self, item)
        else:
            # manual entry
            value_manager = self._manual_values.get(item, None)
            """ :type value_manager: mogwai.properties.base.BaseValueManager | None """
            if value_manager is None:
                raise AttributeError(item)
            return value_manager.value

    def __setitem__(self, key, value):
        logger.debug("Using __setitem__({}): {}...".format(key, value))
        prop = self._properties.get(key, None)
        if prop is not None:
            # call the normal setattr method
            setattr(self, key, value)
        else:
            # manual entry
            if key in self._manual_values:
                # manual entry already exists, update
                self._manual_values[key].setval(value)
            else:
                # manual entry doesn't exist, create
                from mogwai.properties.base import BaseValueManager
                self._manual_values[key] = BaseValueManager(None, value)

    def __delitem__(self, key):
        logger.debug("Using __delitem__({})...".format(key))
        prop = self._properties.get(key, None)
        if prop is not None:
            # call the normal delattr method
            delattr(self, key)
        else:
            # manual entry
            if key in self._manual_values:
                # manual entry already exists, update for delete
                self._manual_values[key] = None
            else:
                raise AttributeError(key)

    def __contains__(self, item):
        logger.debug("Checking contains...")
        return item in set(self._properties.keys()).union(set(self._manual_values.keys()))

    def __len__(self):
        logger.debug("Getting len...")
        return len(set(self._properties.keys()).union(set(self._manual_values.keys())))

    def __iter__(self):
        logger.debug("Iterating through __iter__...")
        for item in set(self._properties.keys()).union(set(self._manual_values.keys())):
            yield item

    def items(self):
        logger.debug("Iterating through items...")
        items = []
        for key in self._properties.keys():
            items.append((key, getattr(self, key)))
        items.extend([(pair[0], pair[1].value) for pair in self._manual_values.items() if pair[1] is not None])
        return items

    def keys(self):
        logger.debug("Iterating through keys...")
        items = []
        items.extend(self._properties.keys())
        items.extend([k for k in self._manual_values.keys() if self._manual_values.get(k) is not None])
        return items

    def values(self):
        logger.debug("Iterating through values...")
        items = []
        for key in self._properties.keys():
            items.append(getattr(self, key))
        items.extend([v.value for v in self._manual_values.values() if v is not None])
        return items


class ElementMetaClass(type):
    """Metaclass for all graph elements"""

    def __new__(mcs, name, bases, body):
        """
        """
        #move graph property definitions into graph property dict
        #and set default column names
        prop_dict = OrderedDict()
        relationship_dict = OrderedDict()

        #get inherited properties
        for base in bases:
            if body.get('__enum_id_only__', None) is None:
                body['__enum_id_only__'] = getattr(base, '__enum_id_only__', True)
            for k, v in getattr(base, '_properties', {}).items():
                prop_dict.setdefault(k, v)
            for k, v in getattr(base, '_relationships', {}).items():
                relationship_dict.setdefault(k, v)

        #print_("Name: %s\n\tBases: %s\n\tBody: %s" % (name, bases, body.keys()))

        def _transform_property(prop_name, prop_obj):
            prop_dict[prop_name] = prop_obj
            prop_obj.set_property_name(prop_name)
            if prop_obj.db_field_prefix is not None:
                db_field_prefix_name = name.lower()
                prop_obj.set_db_field_prefix(db_field_prefix_name)
            #set properties
            _get = lambda self: self._values[prop_name].getval()
            _set = lambda self, val: self._values[prop_name].setval(val)
            _del = lambda self: self._values[prop_name].delval()
            if prop_obj.can_delete:
                body[prop_name] = property(_get, _set, _del)
            else:  # pragma: no cover
                body[prop_name] = property(_get, _set)

        property_definitions = [(k, v) for k, v in body.items() if isinstance(v, properties.GraphProperty)]
        property_definitions = sorted(property_definitions, #cmp=lambda x, y: cmp(x[1].position, y[1].position),
                                      key=lambda x: x[1].position)

        #TODO: check that the defined graph properties don't conflict with any of the
        #Model API's existing attributes/methods transform column definitions
        for k, v in property_definitions:
            _transform_property(k, v)

        #check for duplicate graph property names
        prop_names = set()
        for v in prop_dict.values():
            if v.db_field_name in prop_names:
                raise ModelException("%s defines the graph property %s more than once" % (name, v.db_field_name))
            prop_names.add(v.db_field_name)

        #create db_name -> model name map for loading
        db_map = {}
        for field_name, prop in prop_dict.items():
            db_map[prop.db_field_name] = field_name

        #add management members to the class
        body['_properties'] = prop_dict
        body['_db_map'] = db_map

        ## Manage relationship attributes
        from mogwai.relationships import Relationship
        from mogwai.tools import LazyImportClass
        for k, v in body.items():
            if isinstance(v, (Relationship, LazyImportClass)):
                relationship_dict[k] = v
        body['_relationships'] = relationship_dict

        #auto link gremlin methods
        gremlin_methods = {}

        #get inherited gremlin methods
        for base in bases:
            for k, v in getattr(base, '_gremlin_methods', {}).items():
                gremlin_methods.setdefault(k, v)

        #short circuit __abstract__ inheritance
        body['__abstract__'] = body.get('__abstract__', False)

        #short circuit path inheritance
        gremlin_path = body.get('gremlin_path')
        body['gremlin_path'] = gremlin_path

        def wrap_method(method):
            def method_wrapper(self, *args, **kwargs):
                return method(self, *args, **kwargs)
            return method_wrapper

        for k, v in body.items():
            if isinstance(v, BaseGremlinMethod):
                gremlin_methods[k] = v
                method = wrap_method(v)
                body[k] = method
                if v.classmethod:
                    body[k] = classmethod(method)
                if v.property:
                    body[k] = property(method)

        body['_gremlin_methods'] = gremlin_methods

        #create the class and add a QuerySet to it
        klass = super(ElementMetaClass, mcs).__new__(mcs, name, bases, body)

        #configure the gremlin methods
        for name, method in gremlin_methods.items():
            method.configure_method(klass, name, gremlin_path)

        _add_model_to_space(klass)
        return klass


@add_metaclass(ElementMetaClass)
class Element(BaseElement):
    #__metaclass__ = ElementMetaClass

    @classmethod
    def deserialize(cls, data):
        """ Deserializes rexpro response into vertex or edge objects """

        dtype = data.get('_type')
        data_id = data.get('_id')
        properties = data.get('_properties')
        if dtype == 'vertex':
            vertex_type = properties['element_type']
            if vertex_type not in vertex_types:
                raise ElementDefinitionException('Vertex "%s" not defined' % vertex_type)

            translated_data = vertex_types[vertex_type].translate_db_fields(data)
            return vertex_types[vertex_type](**translated_data)

        elif dtype == 'edge':
            edge_type = data.get('_label') or properties['_label']
            if edge_type not in edge_types:
                raise ElementDefinitionException('Edge "%s" not defined' % edge_type)

            translated_data = edge_types[edge_type].translate_db_fields(data)
            return edge_types[edge_type](data['_outV'], data['_inV'], **translated_data)

        else:
            raise TypeError("Can't deserialize '%s'" % dtype)



from __future__ import unicode_literals
import datetime
from pytz import utc
from nose.plugins.attrib import attr

from mogwai.tests.base import BaseMogwaiTestCase

from mogwai.models import Vertex, Edge, IN, OUT, BOTH, GREATER_THAN, LESS_THAN
from mogwai import properties


# Vertices
class Person(Vertex):
    name = properties.Text()
    age = properties.Integer()


class Course(Vertex):
    name = properties.Text()
    credits = properties.Decimal()


# Edges
class EnrolledIn(Edge):
    date_enrolled = properties.DateTime()
    enthusiasm = properties.Integer(default=5)  # medium, 1-10, 5 by default


class TaughtBy(Edge):
    overall_mood = properties.Text(default='Grumpy')


class BaseTraversalTestCase(BaseMogwaiTestCase):

    @classmethod
    def setUpClass(cls):
        """
        person -enrolled_in-> course
        course -taught_by-> person

        :param cls:
        :return:
        """
        super(BaseTraversalTestCase, cls).setUpClass()
        cls.jon = Person.create(name='Jon', age=143)
        cls.eric = Person.create(name='Eric', age=25)
        cls.blake = Person.create(name='Blake', age=14)

        cls.physics = Course.create(name='Physics 264', credits=1.0)
        cls.beekeeping = Course.create(name='Beekeeping', credits=15.0)
        cls.theoretics = Course.create(name='Theoretical Theoretics', credits=-3.5)

        cls.eric_in_physics = EnrolledIn.create(cls.eric, cls.physics,
                                                date_enrolled=datetime.datetime.now(tz=utc),
                                                enthusiasm=10)  # eric loves physics
        cls.jon_in_beekeeping = EnrolledIn.create(cls.jon, cls.beekeeping,
                                                  date_enrolled=datetime.datetime.now(tz=utc),
                                                  enthusiasm=1)  # jon hates beekeeping

        cls.blake_in_theoretics = EnrolledIn.create(cls.blake, cls.theoretics,
                                                    date_enrolled=datetime.datetime.now(tz=utc),
                                                    enthusiasm=8)

        cls.blake_beekeeping = TaughtBy.create(cls.beekeeping, cls.blake, overall_mood='Pedantic')
        cls.jon_physics = TaughtBy.create(cls.physics, cls.jon, overall_mood='Creepy')
        cls.eric_theoretics = TaughtBy.create(cls.theoretics, cls.eric, overall_mood='Obtuse')

    @classmethod
    def tearDownClass(cls):

        # reverse setup procedure and delete vertices and edges in graph
        cls.eric_theoretics.delete()
        cls.jon_physics.delete()
        cls.blake_beekeeping.delete()
        cls.blake_in_theoretics.delete()
        cls.jon_in_beekeeping.delete()
        cls.eric_in_physics.delete()
        cls.theoretics.delete()
        cls.beekeeping.delete()
        cls.physics.delete()
        cls.blake.delete()
        cls.eric.delete()
        cls.jon.delete()
        super(BaseTraversalTestCase, cls).tearDownClass()


@attr('unit', 'traversals')
class TestVertexTraversals(BaseTraversalTestCase):

    def test_inV_works(self):
        """Test that inV traversals work as expected"""
        results = self.jon.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.physics, results)

        results = self.physics.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.eric, results)

        results = self.eric.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.theoretics, results)

        results = self.theoretics.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.blake, results)

        results = self.beekeeping.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.jon, results)

        results = self.blake.inV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.beekeeping, results)

    def test_inE_traversals(self):
        """Test that inE traversals work as expected"""
        results = self.jon.inE()
        self.assertEqual(len(results), 1)
        self.assertIn(self.jon_physics, results)

    def test_outV_traversals(self):
        """Test that outV traversals work as expected"""
        results = self.eric.outV()
        self.assertEqual(len(results), 1)
        self.assertIn(self.physics, results)

    def test_outE_traverals(self):
        """Test that outE traversals work as expected"""
        results = self.blake.outE()
        self.assertEqual(len(results), 1)
        self.assertIn(self.blake_in_theoretics, results)

    def test_bothE_traversals(self):
        """Test that bothE traversals works"""
        results = self.jon.bothE()
        self.assertEqual(len(results), 2)
        self.assertIn(self.jon_physics, results)
        self.assertIn(self.jon_in_beekeeping, results)

    def test_bothV_traversals(self):
        """Test that bothV traversals work"""
        results = self.blake.bothV()
        self.assertEqual(len(results), 2)
        self.assertIn(self.beekeeping, results)


@attr('unit', 'traversals')
class TestVertexCentricQueries(BaseTraversalTestCase):

    def test_query_vertices(self):
        classes = self.jon.query().labels(EnrolledIn).direction(OUT).vertices()

    def test_query_in(self):
        people = self.physics.query().labels(EnrolledIn).direction(IN).vertices()
        for x in people:
            self.assertIsInstance(x, Person)

    def test_query_out_edges(self):
        classes = self.jon.query().labels(EnrolledIn).direction(OUT).edges()
        for x in classes:
            self.assertIsInstance(x, EnrolledIn, "Expected %s, got %s" % (type(EnrolledIn), type(x)))

    def test_two_labels(self):
        edges = self.jon.query().labels(EnrolledIn, TaughtBy).direction(BOTH).edges()
        for e in edges:
            self.assertIsInstance(e, (EnrolledIn, TaughtBy))

    def test_has(self):
        self.assertEqual(0, len(self.jon.query().labels(EnrolledIn).has('enrolledin_enthusiasm', 5,
                                                                        GREATER_THAN).vertices()))
        num = self.jon.query().labels(EnrolledIn).has('tests_vertex_traversals_tests_enthusiasm', 5, GREATER_THAN).count()
        self.assertEqual(0, num)

        self.assertEqual(1, len(self.jon.query().labels(EnrolledIn).has('enrolledin_enthusiasm', 5,
                                                                        LESS_THAN).vertices()))
        num = self.jon.query().labels(EnrolledIn).has('enrolledin_enthusiasm', 5, LESS_THAN).count()
        self.assertEqual(1, num)

    def test_interval(self):
        self.assertEqual(1, len(self.blake.query().labels(EnrolledIn).interval('enrolledin_enthusiasm', 2,
                                                                               9).vertices()))
        self.assertEqual(1, len(self.blake.query().labels(EnrolledIn).interval('enrolledin_enthusiasm', 9,
                                                                               2).vertices()))
        self.assertEqual(0, len(self.blake.query().labels(EnrolledIn).interval('enrolledin_enthusiasm', 2,
                                                                               8).vertices()))



from __future__ import unicode_literals
from mogwai._compat import print_
from mogwai.metrics.base import BaseMetricsReporter
import newrelic


class NewRelicReporter(BaseMetricsReporter):
    """ NewRelic Metrics Reporter class

    If you have a hosted newrelic service, you can log the OGM metrics to your newrelic service.
    """

    def __init__(self, config_file, environment=None, *args, **kwargs):
        """ Create a Metrics Reporter

        To override this method, please make sure to call the superclass *after* your methods. Ie.
        def __init__(self, mystuff, stuff2=None, *args, **kwargs):
            # do my stuff...
            super(MyClass, self).__init__(*args, **kwargs)

        :param config_file: The NewRelic Configuration File
        :type config_file: basestring
        :param environment: The NewRelic Configuration Environment (optional)
        :type environment: basestring
        :param metric_prefix: The prefix on the collected metrics. The default implementation ignores this.
        :type metric_prefix: basestring | None
        :param registry: The pyformance registry
        :type registry: list[ MetricsRegistry | RegexRegistry ] | MetricsRegistry | RegexRegistry
        :param reportingInterval: The interval (number of seconds) on which to report the collected metrics.
        :type reportingInterval: float | long | int
        """
        self.agent = newrelic.agent
        self.agent.initialize(config_file, environment=environment)
        super(NewRelicReporter, self).__init__(*args, **kwargs)

    def convert_metric_name(self, key, metric):
        """ Converts the mogwai naming schema to NewRelic naming schema

        :param key: the mogwai metric key
        :type key: basestring
        :param metric: the specific metric being collected
        :type key: basestring
        :returns: Converted key
        :rtype: basestring
        """
        return 'Custom/' + key.replace('.', '/') + '/' + metric

    def send_metrics(self):
        """ Default Hosted Graphite implementation is to dump all metrics to STDOUT

        Change this if you want to customize for a particular metric collection system, ie. graphite, newrelic, etc.
        """
        metrics = self.get_metrics()
        if not metrics:
            return

        for mkey, metric in metrics.items():
            for mname, mval in metric.items():
                try:
                    self.agent.record_custom_metric(self.convert_metric_name(mkey, mname), mval, None)
                except Exception as e:
                    print_(e)
